"""SCM Role."""
from scm_helper.coach import check_coach_permissions
from scm_helper.config import (
    A_GUID,
    A_ISVOLUNTEER,
    A_MEMBERS,
    C_CHECK_PERMISSIONS,
    C_CHECK_RESTRICTIONS,
    C_IS_COACH,
    C_LOGIN,
    C_MANDATORY,
    C_ROLE,
    C_ROLES,
    C_UNUSED,
    C_VOLUNTEER,
    get_config,
)
from scm_helper.entity import Entities, Entity
from scm_helper.issue import (
    E_COACH_ROLE,
    E_INACTIVE,
    E_NO_LOGIN,
    E_NO_RESTRICTIONS,
    E_NO_SWIMMERS,
    E_UNUSED_LOGIN,
    E_VOLUNTEER,
    debug_trace,
    issue,
)


class Roles(Entities):
    """Roles."""

    def new_entity(self, entity):
        """Create a new entity."""
        return Role(entity, self.scm, self._url)


class Role(Entity):
    """A role."""

    @debug_trace(5)
    def analyse(self):
        """Analyse the role."""
        cfg = get_config(self.scm, C_ROLES, C_ROLE)
        unused = get_config(self.scm, C_ROLES, C_LOGIN, C_UNUSED)

        if len(self.members) == 0:
            issue(self, E_NO_SWIMMERS, "Role")
            return

        for member in self.members:
            self.check_role_member(member, unused)

            if cfg and (self.name in cfg):
                self.check_role_permissions(member)

    def check_role_member(self, member, unused):
        """Check out a role member."""
        if member.is_active is False:
            issue(member, E_INACTIVE, f"Member of role {self.name} (fixable)")
            if self.newdata and (A_MEMBERS in self.newdata):
                fix = self.newdata
            else:
                fix = {}
                fix[A_MEMBERS] = self.data[A_MEMBERS].copy()
            fix[A_MEMBERS].remove({A_GUID: member.guid})
            self.fixit(fix, f"Delete from role {self.name}")

        if member.username is None:
            issue(member, E_NO_LOGIN, f"Member of role {self.name}, so cannot login")

        if get_config(self.scm, C_ROLES, C_ROLE, self.name, C_IS_COACH):
            if member.is_coach is False:
                issue(member, E_COACH_ROLE, f"Role: {self.name}")

        if get_config(self.scm, C_ROLES, C_VOLUNTEER, C_MANDATORY):
            if member.is_volunteer is False:
                issue(member, E_VOLUNTEER, f"Role: {self.name} (fixable)")
                fix = {}
                fix[A_ISVOLUNTEER] = "1"
                member.fixit(fix, "Mark as volunteer")

        if member.last_login:
            if (self.scm.today - member.last_login).days > unused:
                issue(
                    member,
                    E_UNUSED_LOGIN,
                    f"Role: {self.name} [Last login: {member.last_login}]",
                )
        else:
            issue(member, E_UNUSED_LOGIN, f"Role: {self.name} [Never]")

    def check_role_permissions(self, member):
        """Check out a role permissions."""
        lookup = get_config(self.scm, C_ROLES, C_ROLE, self.name)
        if lookup is None:
            return

        if (C_CHECK_PERMISSIONS in lookup) and lookup[C_CHECK_PERMISSIONS]:
            check_coach_permissions(member, self)

        if (C_CHECK_RESTRICTIONS in lookup) and lookup[C_CHECK_RESTRICTIONS]:
            if len(member.restricted) == 0:
                issue(member, E_NO_RESTRICTIONS, f"Role: {self.name}")

    @property
    def name(self):
        """Guid."""
        return self.data["RoleName"]



"""Configuration stuff."""
from schema import And, Optional, Regex, Schema, SchemaError

from scm_helper.notify import notify
from scm_helper.version import VERSION

# SCM access URLs etc
URL_BASE = "https://api.swimclubmanager.co.uk"

URL_CONDUCT = f"{URL_BASE}/CodeOfConduct"
URL_EVENTS = f"{URL_BASE}/ClubEvents"
URL_GROUPS = f"{URL_BASE}/ClubGroups"
URL_INCIDENTBOOK = f"{URL_BASE}/IncidentBook"
URL_LISTS = f"{URL_BASE}/EmailLists"
URL_MEETS = f"{URL_BASE}/Meets"
URL_MEMBERS = f"{URL_BASE}/Members"
URL_NOTICE = f"{URL_BASE}/NoticeBoard"
URL_ROLES = f"{URL_BASE}/ClubRoles"
URL_SESSIONS = f"{URL_BASE}/ClubSessions"
URL_TRIALS = f"{URL_BASE}/TrialRequests"
URL_WAITINGLIST = f"{URL_BASE}/WaitingList"
URL_WHO = f"{URL_BASE}/WhosWho"

HELPURL = "https://github.com/ColinRobbins/scm-helper/wiki"

USER_AGENT = f"SCM-Helper-v{VERSION} ###CLUB_NAME###"

# Do not change below here...

BACKUP_DIR = "backups"
CONFIG_DIR = "scm-helper"
CONFIG_FILE = "config.yaml"
KEYFILE = "apikey.enc"
RECORDS_DIR = "records"

CODES_OF_CONDUCT = "Conduct"
EVENTS = "Club Events"
GROUPS = "Groups"
INCIDENTBOOK = "Incident Book"
LISTS = "Lists"
MEETS = "Meets"
MEMBERS = "Members"
NOTICE = "Notice Board"
ROLES = "Roles"
SESSIONS = "Sessions"
TRIALS = "Trial Requests"
WAITINGLIST = "Waiting List"
WHO = "Whos Who"

BACKUP_URLS = [
    [INCIDENTBOOK, URL_INCIDENTBOOK],
    [EVENTS, URL_EVENTS],
    [MEETS, URL_MEETS],
    [TRIALS, URL_TRIALS],
    [WAITINGLIST, URL_WAITINGLIST],
    [NOTICE, URL_NOTICE],
]


FILE_READ = "r"
FILE_WRITE = "w"

# JSON labels
GUID = "Guid"

# Exceptions in Notes
EXCEPTION_EMAILDIFF = "API: different email OK"
EXCEPTION_GENERAL = "API: Exception"
EXCEPTION_GROUPNOSESSION = "API: no sessions OK"
EXCEPTION_NODBS = "API: Coach no DBS OK"
EXCEPTION_NOEMAIL = "API: no email OK"
EXCEPTION_NOGROUPS = "API: no groups OK"
EXCEPTION_NONSWIMMINGMASTER = "API: non swimming master"
EXCEPTION_NOSAFEGUARD = "API: Coach no Safeguard OK"
EXCEPTION_NOSESSIONS = "API: Coach no sessions"
EXCEPTION_PERMISSIONS = "API: Coach permission OK"
EXCEPTION_TWOGROUPS = "API: two groups OK"
EXCEPTION_SE_HIDDEN = "API: Hidden in SE Database"
EXCEPTION_SE_NAME = "API: Name in SE Database OK"

SCM_DATE_FORMAT = "%Y-%m-%d"
PRINT_DATE_FORMAT = "%d-%m-%Y"
SCM_CSV_DATE_FORMAT = "%d/%m/%Y"

# SCM Data Attributes
A_ACTIVE = "Active"
A_ARCHIVED = "Archived"
A_ASA_CATEGORY = "ASACategory"
A_ASA_NUMBER = "ASANumber"
A_DATEAGREED = "DateAgreed"
A_DOB = "DOB"
A_FIRSTNAME = "Firstname"
A_GUID = "Guid"
A_ISCOACH = "IsACoach"
A_ISPARENT = "IsAParent"
A_ISVOLUNTEER = "IsAVolunteer"
A_KNOWNAS = "KnownAs"
A_LAST_ATTENDED = "LastAttended"
A_LASTNAME = "Lastname"
A_MAX_MEMBERS = "MaxMembers"
A_MEMBERS = "Members"
A_PARENTS = "Parents"
A_USERNAME = "Username"

# Config parameters
C_ABSENCE = "absence"
C_AGE = "age"
C_AGE_EOY = "age_eoy"
C_ALIGN_QUARTER = "align_quarter"
C_ALLOW_GROUP = "allow_group"
C_ALLOW_UPDATE = "allow_update"
C_ALL_AGES = "all_ages_u18"
C_BASE_URL = "base_url"
C_BROWSER = "browser"
C_CHECK_DBS = "check_dbs"
C_CHECK_PERMISSIONS = "check_permissions"
C_CHECK_RESTRICTIONS = "check_restrictions"
C_CHECK_SE_NUMBER = "check_se_number"
C_CHECK_URL = "check_url"
C_CHILD = "child"
C_CLASS = "class"
C_CLUB = "club"
C_COACH = "coach"
C_COACHES = "coaches"
C_CONDUCT = "conduct"
C_CONF_DIFF = "confirmation_difference"
C_CONFIRMATION = "confirmation"
C_COVID = "covid"
C_DATE = "date"
C_DBS = "dbs"
C_DEBUG_LEVEL = "debug_level"
C_DOB_FORMAT = "dob_format"
C_EDIT = "edit"
C_EMAIL = "email"
C_EXPIRY = "expiry"
C_FACEBOOK = "facebook"
C_FILES = "files"
C_GENDER = "gender"
C_GRACE = "grace"
C_GROUP = "group"
C_GROUPS = "groups"
C_IGNORE = "ignore"
C_IGNORE_ATTENDANCE = "ignore_attendance"
C_IGNORE_COACH = "ignore_coach"
C_IGNORE_COMMITTEE = "ignore_committee"
C_IGNORE_ERROR = "ignore_error"
C_IGNORE_GROUP = "ignore_group"
C_IGNORE_SWIMMER = "ignore_swimmer"
C_IGNORE_UNKNOWN = "ignore_unknown"
C_INACTIVE = "inactive"
C_IS_COACH = "is_coach"
C_ISSUES = "issues"
C_JOBTITLE = "jobtitle"
C_LIST = "list"
C_LISTS = "lists"
C_LOGIN = "login"
C_MANDATORY = "mandatory"
C_MAPPING = "mapping"
C_MAX_AGE = "max_age"
C_MAX_AGE_EOY = "max_age_eoy"
C_MAX_SESSIONS = "max_sessions"
C_MAX_YEAR = "max_year"
C_MEMBERS = "members"
C_MESSAGE = "message"
C_MIN_AGE = "min_age"
C_MIN_AGE_EOY = "min_age_eoy"
C_MIN_YEAR = "min_year"
C_NAME = "name"
C_NEWSTARTER = "newstarter"
C_NO_CLUB_SESSIONS = "no_club_sessions"
C_NO_SESSION_ALLOWED = "no_session_allowed"
C_NO_SESSIONS = "no_sessions"
C_OVERALL_FASTEST = "overall_fastest"
C_PARENT = "parent"
C_PARENTS = "parents"
C_PASSWORD = "password"
C_PRIORITY = "priority"
C_RECORDS = "records"
C_REGISTER = "register"
C_RELAY = "relay"
C_ROLE = "role"
C_ROLES = "roles"
C_SELENIUM = "selenium"
C_SEND_TO = "send_to"
C_SE_ONLY = "se_only"
C_SESSION = "session"
C_SESSIONS = "sessions"
C_SMTP_PORT = "smtp_port"
C_SMTP_SERVER = "smtp_server"
C_SUFFIX = "suffix"
C_SWIM_ENGLAND = "swim_england"
C_SWIMMER = "swimmer"
C_SWIMMERS = "swimmers"
C_TEST_ID = "test_id"
C_TIME = "time"
C_TLS = "tls"
C_TYPE = "type"
C_TYPES = "types"
C_UNIQUE = "unique"
C_UNUSED = "unused"
C_USERNAME = "username"
C_VERIFY = "verify"
C_VOLUNTEER = "volunteer"
C_WEB_DRIVER = "web_driver"

CTYPE_COACH = "coach"
CTYPE_COMMITTEE = "committee"
CTYPE_LIFESAVER = "lifesaver"
CTYPE_LTS = "lts_teacher"
CTYPE_OPENWATER = "openwater"
CTYPE_PARENT = "parent"
CTYPE_POLO = "waterpolo"
CTYPE_SWIMMER = "swimmer"
CTYPE_SYNCHRO = "synchro"
CTYPE_VOLUNTEER = "volunteer"

CTYPES = [
    CTYPE_COACH,
    CTYPE_COMMITTEE,
    CTYPE_LIFESAVER,
    CTYPE_LTS,
    CTYPE_OPENWATER,
    CTYPE_PARENT,
    CTYPE_POLO,
    CTYPE_SWIMMER,
    CTYPE_SYNCHRO,
    CTYPE_VOLUNTEER,
]

O_NEWSTARTER = "--newstarter"
O_TO = "--to"
O_VERIFY = "--verify"
O_BACKUP = "--backup"
O_FORMAT = "--format"
O_FIX = "--fix"

VAR_CONDUCT = []
VAR_GROUP = []
VAR_ISSUE = []
VAR_ROLE = []
VAR_SESSION = []


def group(data):
    """Register group."""
    VAR_GROUP.append(data)
    return True


def session(data):
    """Register session."""
    VAR_SESSION.append(data)
    return True


def conduct(data):
    """Register conduct."""
    VAR_CONDUCT.append(data)
    return True


def role(data):
    """Register role."""
    VAR_ROLE.append(data)
    return True


def issue(data):
    """Register issues."""
    VAR_ISSUE.append(data)
    return True


def member_type(data):
    """check_type."""
    if data in CTYPES:
        return True
    return False


SCHEMA = Schema(
    {
        C_CLUB: str,
        C_ALLOW_UPDATE: bool,
        Optional(C_DEBUG_LEVEL): int,
        Optional(C_EMAIL): {
            C_USERNAME: str,
            C_SMTP_SERVER: str,
            C_SMTP_PORT: int,
            C_SEND_TO: str,
            C_TLS: bool,
            C_PASSWORD: str,
        },
        C_SWIMMERS: {
            C_USERNAME: {C_MIN_AGE: int},
            C_PARENT: {C_MANDATORY: bool, C_MAX_AGE: int},
            Optional(C_CONF_DIFF): {C_VERIFY: bool},
            Optional(C_ABSENCE): {C_TIME: int},
            Optional(C_MAX_SESSIONS): int,
        },
        C_PARENTS: {
            Optional(C_AGE): {C_MIN_AGE: int, C_CHILD: int},
            Optional(C_LOGIN): {C_MANDATORY: bool},
        },
        C_MEMBERS: {
            Optional(C_CONFIRMATION): {C_EXPIRY: int, Optional(C_ALIGN_QUARTER): bool},
            Optional(C_DBS): {C_EXPIRY: int},
            Optional(C_NEWSTARTER): {C_GRACE: int},
            Optional(C_INACTIVE): {C_TIME: int},
        },
        C_COACHES: {C_ROLE: {C_MANDATORY: bool}},
        Optional(C_ROLES): {
            Optional(C_VOLUNTEER): {C_MANDATORY: bool},
            Optional(C_LOGIN): {C_UNUSED: int},
            Optional(C_ROLE): {
                role: {
                    Optional(C_CHECK_PERMISSIONS): bool,
                    Optional(C_IS_COACH): bool,
                    Optional(C_CHECK_RESTRICTIONS): bool,
                }
            },
        },
        Optional(C_GROUPS): {
            Optional(C_PRIORITY): [group],
            Optional(C_GROUP): {
                group: {
                    Optional(C_CHECK_DBS): bool,
                    Optional(C_CONFIRMATION): str,
                    Optional(C_IGNORE_GROUP): bool,
                    Optional(C_IGNORE_SWIMMER): bool,
                    Optional(C_IGNORE_UNKNOWN): bool,
                    Optional(C_MAX_AGE): int,
                    Optional(C_MIN_AGE): int,
                    Optional(C_NO_CLUB_SESSIONS): bool,
                    Optional(C_NO_SESSION_ALLOWED): [group],
                    Optional(C_NO_SESSIONS): bool,
                    Optional(C_SESSIONS): [session],
                    Optional(C_TYPE): member_type,
                    Optional(C_UNIQUE): bool,
                }
            },
        },
        Optional(C_SESSIONS): {
            Optional(C_ABSENCE): int,
            Optional(C_REGISTER): int,
            Optional(C_COVID): conduct,
            Optional(C_SESSION): {
                session: {
                    Optional(C_GROUPS): [group],
                    Optional(C_IGNORE_ATTENDANCE): bool,
                }
            },
        },
        Optional(C_CONDUCT): {
            conduct: {
                C_TYPES: [member_type],
                Optional(C_IGNORE_GROUP): [group],
                Optional(C_DATE): Regex(r"\d\d\d\d-\d\d-\d\d"),
            }
        },
        Optional(C_ISSUES): {
            issue: {Optional(C_MESSAGE): str, Optional(C_IGNORE_ERROR): bool}
        },
        Optional(C_JOBTITLE): {C_IGNORE: [str]},
        Optional(C_FILES): {
            str: {
                Optional(C_CHECK_SE_NUMBER): bool,
                Optional(C_IGNORE_GROUP): [group],
                C_MAPPING: {
                    A_FIRSTNAME: str,
                    A_LASTNAME: str,
                    Optional(A_KNOWNAS): str,
                    Optional(A_ASA_NUMBER): str,
                    Optional(A_ASA_CATEGORY): str,
                    Optional(A_DOB): str,
                    Optional(C_DOB_FORMAT): str,
                },
            }
        },
        Optional(C_TYPES): {
            member_type: {
                Optional(C_CHECK_SE_NUMBER): bool,
                Optional(C_IGNORE_COACH): bool,
                Optional(C_IGNORE_COMMITTEE): bool,
                Optional(C_NAME): str,
                Optional(C_JOBTITLE): bool,
                Optional(C_GROUPS): [group],
                Optional(C_PARENTS): bool,
            }
        },
        Optional(C_LISTS): {
            C_SUFFIX: str,
            C_EDIT: bool,
            C_CONFIRMATION: bool,
            Optional(C_CONDUCT): [conduct],
            Optional(C_LIST): {
                str: {
                    Optional(C_GENDER): And(str, lambda s: s in ("male", "female")),
                    Optional(C_GROUP): group,
                    Optional(C_ALLOW_GROUP): group,
                    Optional(C_UNIQUE): bool,
                    Optional(C_MAX_AGE): int,
                    Optional(C_MAX_AGE_EOY): int,
                    Optional(C_MAX_YEAR): int,
                    Optional(C_MIN_AGE): int,
                    Optional(C_MIN_AGE_EOY): int,
                    Optional(C_MIN_YEAR): int,
                    Optional(C_TYPE): member_type,
                }
            },
        },
        Optional(C_FACEBOOK): {Optional(C_FILES): [str], Optional(C_GROUPS): [str]},
        Optional(C_SELENIUM): {C_BROWSER: str, C_WEB_DRIVER: str},
        Optional(C_SWIM_ENGLAND): {C_BASE_URL: str, C_CHECK_URL: str, C_TEST_ID: int},
        Optional(C_RECORDS): {
            Optional(C_RELAY): bool,
            Optional(C_AGE_EOY): bool,
            Optional(C_VERIFY): bool,
            Optional(C_SE_ONLY): bool,
            Optional(C_ALL_AGES): bool,
            Optional(C_OVERALL_FASTEST): bool,
        },
    }
)


def get_config(scm, item, item1=None, item2=None, item3=None):
    """Get config parameters..."""
    # pylint: disable=too-many-return-statements
    # Yes, need them all...

    cfg = scm.config(item)
    if cfg is None:
        return None

    if item1 is None:
        return cfg

    if item1 not in cfg:
        return None

    if item2 is None:
        return cfg[item1]

    if item2 not in cfg[item1]:
        return None

    if item3 is None:
        return cfg[item1][item2]

    if item3 not in cfg[item1][item2]:
        return None

    return cfg[item1][item2][item3]


def verify_schema(data):
    """Verify the config file."""
    try:
        SCHEMA.validate(data)
        return True
    except SchemaError as error:
        notify(f"Error in config file:\n{error}\n")
        return False


def verify_schema_data(scm):
    """Verify the data in the schema."""
    error = False
    for xgroup in VAR_GROUP:
        if xgroup not in scm.groups.by_name:
            notify(f"Error in config file: Group '{xgroup}' not found\n")
            error = True

    for code in VAR_CONDUCT:
        if code not in scm.conduct.by_name:
            notify(f"Error in config file: Code of Conduct '{code}' not found\n")
            error = True

    for xrole in VAR_ROLE:
        if xrole not in scm.roles.by_name:
            notify(f"Error in config file: Role '{xrole}' not found\n")
            error = True

    for xsession in VAR_SESSION:
        if scm.sessions.find_session_substr(xsession) is None:
            notify(f"Error in config file: Session '{xsession}' not found\n")
            error = True

    for xissue in VAR_ISSUE:
        if scm.issue_handler.check_issue(xissue) is False:
            notify(f"Error in config file: Issue '{xissue}' not found\n")
            error = True

    if error:
        return False
    return True


def check_default(scm):
    """Give warning if config not made."""

    msg = ""
    if get_config(scm, C_ROLES, C_ROLE) is None:
        msg += " - No Roles configured\n"
    if get_config(scm, C_GROUPS) is None:
        msg += " - No Groups configured\n"
    if get_config(scm, C_SESSIONS, C_SESSION) is None:
        msg += " - No Sessions configured\n"
    if get_config(scm, C_CONDUCT) is None:
        msg += " - No Code of Conduct configured\n"
    if get_config(scm, C_LISTS, C_LIST) is None:
        msg += " - No Lists configured\n"

    if msg:
        msg = "\nIn your configuration file, you have:\n" + msg
        msg += "By configuring these, SCM Helper can do a more for you!\n\n"

    return msg



"""SCM List."""
from scm_helper.config import (
    A_GUID,
    A_MEMBERS,
    C_ALLOW_GROUP,
    C_EDIT,
    C_GENDER,
    C_GROUP,
    C_LIST,
    C_LISTS,
    C_MAX_AGE,
    C_MAX_AGE_EOY,
    C_MAX_YEAR,
    C_MIN_AGE,
    C_MIN_AGE_EOY,
    C_MIN_YEAR,
    C_SUFFIX,
    C_TYPE,
    C_UNIQUE,
    EXCEPTION_NOEMAIL,
    get_config,
)
from scm_helper.entity import Entities, Entity, check_type
from scm_helper.issue import E_LIST_ERROR, E_NO_SWIMMERS, debug_trace, issue
from scm_helper.notify import notify

A_LISTNAME = "ListName"


class Lists(Entities):
    """Lists."""

    def __init__(self, scm, name, url):
        """Initilaise."""
        super().__init__(scm, name, url)
        self._suffix = None
        self.by_name = {}
        self.newlists = []

    def new_entity(self, entity):
        """Create a new entity for list in SCM."""
        xlist = List(entity, self.scm, self._url)
        self.by_name[xlist.name] = xlist

        return xlist

    def update(self):
        """Create a new list to add to SCM."""
        cfg = get_config(self.scm, C_LISTS)
        if cfg is None:
            return

        if get_config(self.scm, C_LISTS, C_EDIT) is not True:
            notify("List update prohibited by config.\n")
            return

        lists = get_config(self.scm, C_LISTS, C_LIST)
        self._suffix = get_config(self.scm, C_LISTS, C_SUFFIX)

        if lists:
            for xlist in lists:
                newlist = NewList(self.scm, xlist, self._url)
                self.newlists.append(newlist)
                newlist.populate()

        # Separate for loop, as add_to_list may have created some too
        for xlist in self.newlists:
            xlist.generate_data(self._suffix)

            if xlist.upload() is None:
                pass  # not sure what to do, just carry on!

    def delete(self):
        """Delete all members."""
        super().delete()
        for entity in self.newlists:
            entity.delete()
            del entity
        self.newlists = []

    def add(self, name, person):
        """Add a person to a new list."""
        for xlist in self.newlists:
            if xlist.name == name:
                xlist.add_member(person)
                return

        newlist = NewList(self.scm, name, self._url)
        self.newlists.append(newlist)
        newlist.add_member(person)


class List(Entity):
    """An existing list."""

    @debug_trace(5)
    def analyse(self):
        """Analyse existing lists."""
        if len(self.members) == 0:
            issue(self, E_NO_SWIMMERS, "List")
            return

        for member in self.members:
            if member.is_active is False:
                # Never get here as entity linkage prevents it.
                msg = f"Inactive but on email list {self.name} (fixable)"
                issue(member, E_LIST_ERROR, msg)
                if self.newdata and (A_MEMBERS in self.newdata):
                    fix = self.newdata
                else:
                    fix = {}
                    fix[A_MEMBERS] = self.data[A_MEMBERS].copy()
                fix[A_MEMBERS].remove({A_GUID: member.guid})
                self.fixit(fix, f"Delete {member.name}")

            if member.email is None:
                issue(member, E_LIST_ERROR, f"No email, but on email list {self.name}")

    @property
    def name(self):
        """Guid."""
        return self.data["ListName"]


class NewList(Entity):
    """A list."""

    # pylint: disable=too-many-instance-attributes

    def __init__(self, scm, xlist, url):
        """Initilaise."""
        # pylint: disable=super-init-not-called
        # Do not call super...
        self.data = None
        self.members = []
        self._scm = scm
        self.ignore = False
        self._name = xlist
        self.new_list = True
        self.newdata = {}
        self.url = url

    def populate(self):
        """Search for entries and fill the list."""
        # pylint: disable=too-many-branches, too-many-statements
        cfg = get_config(self.scm, C_LISTS, C_LIST, self.name)

        # set defaults
        min_age = 0
        if C_MIN_AGE in cfg:
            min_age = cfg[C_MIN_AGE]

        max_age = 999
        if C_MAX_AGE in cfg:
            max_age = cfg[C_MAX_AGE]

        min_age_eoy = 0
        if C_MIN_AGE_EOY in cfg:
            min_age_eoy = cfg[C_MIN_AGE_EOY]

        max_age_eoy = 999
        if C_MAX_AGE_EOY in cfg:
            max_age_eoy = cfg[C_MAX_AGE_EOY]

        min_year = 1900
        if C_MIN_YEAR in cfg:
            min_year = cfg[C_MIN_YEAR]

        max_year = 2200
        if C_MAX_YEAR in cfg:
            max_year = cfg[C_MAX_YEAR]

        for member in self.scm.members.entities:

            if member.is_active is False:
                continue

            if member.in_ignore_group:
                continue

            if member.age and (member.age < min_age):
                continue

            if member.age and (member.age > max_age):
                continue

            if member.age_eoy and (member.age_eoy < min_age_eoy):
                continue

            if member.age_eoy and (member.age_eoy > max_age_eoy):
                continue

            if member.dob and (member.dob.year > max_year):
                continue

            if member.dob and (member.dob.year < min_year):
                continue

            if C_GROUP in cfg:
                if member.find_group(cfg[C_GROUP]) is False:
                    continue
                if C_UNIQUE in cfg:
                    if len(member.groups) > 1:
                        if C_ALLOW_GROUP not in cfg:
                            continue
                        if member.find_group(cfg[C_ALLOW_GROUP]) is False:
                            continue

            if C_GENDER in cfg:
                gender = cfg[C_GENDER]
                xgender = "F"
                if gender == "male":
                    xgender = "M"
                if member.gender != xgender:
                    continue

            if C_TYPE in cfg:
                xtype = cfg[C_TYPE]
                if check_type(member, xtype) is False:
                    continue

            if member.email is None:
                if member.print_exception(EXCEPTION_NOEMAIL):
                    msg = f"No email, but required for email list {self.name}"
                    issue(member, E_LIST_ERROR, msg)
                continue

            self.add_member(member)

    def generate_data(self, suffix):
        """Create data to upload."""
        listname = f"{self.name}{suffix}"
        self.newdata[A_LISTNAME] = listname
        self.newdata[A_MEMBERS] = []
        for member in self.members:
            self.newdata[A_MEMBERS].append({A_GUID: member})

        xlist = None
        if listname in self.scm.lists.by_name:
            xlist = self.scm.lists.by_name[listname]
            self.newdata[A_GUID] = xlist.guid
            self.new_list = False

    def upload(self):
        """Create data to upload."""
        notify(f"Creating / Updating list: {self.name}\n")
        return self.scm.api_write(self, self.new_list)

    def add_member(self, member):
        """Add a member to the list."""
        if member.email:
            self.members.append(member.guid)

    @property
    def name(self):
        """name."""
        return self._name



"""Swimmer routines."""

from scm_helper.config import (
    C_ABSENCE,
    C_CHECK_SE_NUMBER,
    C_CONF_DIFF,
    C_GROUP,
    C_GROUPS,
    C_IGNORE_ATTENDANCE,
    C_MANDATORY,
    C_MAX_AGE,
    C_MAX_SESSIONS,
    C_MIN_AGE,
    C_NO_CLUB_SESSIONS,
    C_PARENT,
    C_PARENTS,
    C_SESSION,
    C_SESSIONS,
    C_SWIMMERS,
    C_TIME,
    C_TYPES,
    C_UNIQUE,
    C_USERNAME,
    C_VERIFY,
    CTYPE_POLO,
    CTYPE_SYNCHRO,
    EXCEPTION_EMAILDIFF,
    EXCEPTION_NOGROUPS,
    EXCEPTION_TWOGROUPS,
    PRINT_DATE_FORMAT,
    get_config,
)
from scm_helper.issue import (
    E_ABSENT,
    E_ASA,
    E_CONFIRM_DIFF,
    E_DATE_JOINED,
    E_DOB,
    E_EMAIL_MATCH,
    E_GENDER,
    E_INACTIVE,
    E_LOGIN_TOO_YOUNG,
    E_MAX_SESSIONS,
    E_NEVERSEEN,
    E_NO_GROUP,
    E_NO_PARENT,
    E_NUM_PARENTS,
    E_TWO_GROUPS,
    debug,
    issue,
)


def analyse_swimmer(swimmer):
    """Analyse a swimmer..."""
    # pylint: disable=too-many-branches
    if swimmer.in_ignore_group:
        return

    if swimmer.in_ignore_swimmer:
        return

    if len(swimmer.groups) == 0:
        if swimmer.print_exception(EXCEPTION_NOGROUPS):
            issue(swimmer, E_NO_GROUP)

    if swimmer.dob is None:
        issue(swimmer, E_DOB)

    if swimmer.gender is None:
        issue(swimmer, E_GENDER)

    if swimmer.date_joined is None:
        issue(swimmer, E_DATE_JOINED)

    check_asa(swimmer)
    check_lastseen(swimmer)
    check_two_groups(swimmer)
    check_login(swimmer)

    if swimmer.is_swimmer:
        check_parents(swimmer)
        check_max_sessions(swimmer)
        return

    if swimmer.is_synchro:
        if get_config(swimmer.scm, C_TYPES, CTYPE_SYNCHRO, C_PARENTS) is False:
            pass
        else:
            check_parents(swimmer)
            return

    if swimmer.is_polo:
        if get_config(swimmer.scm, C_TYPES, CTYPE_POLO, C_PARENTS) is False:
            pass
        else:
            check_parents(swimmer)


def check_max_sessions(swimmer):
    """Check not exceeding max numbr of sessions allowed"""
    max_session = get_config(swimmer.scm, C_SWIMMERS, C_MAX_SESSIONS)
    if max_session:
        if len(swimmer.sessions) > max_session:
            sessions = swimmer.print_swimmer_sessions(False)
            issue(swimmer, E_MAX_SESSIONS, f"{len(swimmer.sessions)}: \n{sessions}")


def check_asa(swimmer):
    """Check ASA (Swim England) number is OK."""
    if swimmer.asa_number is None:
        cfg_synchro = get_config(swimmer.scm, C_TYPES, CTYPE_SYNCHRO, C_CHECK_SE_NUMBER)
        cfg_polo = get_config(swimmer.scm, C_TYPES, CTYPE_POLO, C_CHECK_SE_NUMBER)

        err = True
        if swimmer.is_polo and (cfg_polo is False):
            err = False
        if swimmer.is_synchro and (cfg_synchro is False):
            err = False
        if err:
            issue(swimmer, E_ASA)


def check_lastseen(swimmer):
    """Check when swimemr was last seen."""
    if swimmer.lastseen is None:
        if len(swimmer.sessions) > 0:
            # if no session, don't have data about when they have been seen
            check = True
            for session in swimmer.sessions:
                if get_config(
                    swimmer.scm,
                    C_SESSIONS,
                    C_SESSION,
                    session.name,
                    C_IGNORE_ATTENDANCE,
                ):
                    check = False
                    continue
            if check:
                issue(swimmer, E_NEVERSEEN)
    else:
        gap = (swimmer.scm.today - swimmer.lastseen).days
        absence = get_config(swimmer.scm, C_SWIMMERS, C_ABSENCE, C_TIME)
        if gap > absence:
            when = swimmer.lastseen.strftime(PRINT_DATE_FORMAT)
            issue(swimmer, E_ABSENT, f"Last seen: {when}")


def check_login(swimmer):
    """Check if the login is OK."""
    if swimmer.username:
        min_age = get_config(swimmer.scm, C_SWIMMERS, C_USERNAME, C_MIN_AGE)
        if min_age:
            if swimmer.age and (swimmer.age < min_age):
                issue(swimmer, E_LOGIN_TOO_YOUNG, f"Age: {swimmer.age}")


def check_two_groups(swimmer):
    """Check if swimmer in two groups."""
    if get_config(swimmer.scm, C_GROUPS, C_GROUP) is None:
        return  # No config, so ignore error.

    g_count = 0
    errmsg = ""
    for group in swimmer.groups:

        nosession = get_config(
            swimmer.scm, C_GROUPS, C_GROUP, group.name, C_NO_CLUB_SESSIONS
        )
        if nosession is True:
            continue

        unique = get_config(swimmer.scm, C_GROUPS, C_GROUP, group.name, C_UNIQUE)
        if unique is None:
            unique = True

        if unique:
            g_count += 1
            if g_count > 1:
                errmsg += f", {group.name}"
            else:
                errmsg += group.name

    if g_count > 1:
        if swimmer.print_exception(EXCEPTION_TWOGROUPS):
            issue(swimmer, E_TWO_GROUPS, errmsg)


def check_parents(swimmer):
    """Check consistence between swimmer and parent email."""
    # pylint: disable=too-many-branches
    email = None
    match = False
    count = 0
    confirm_error = False

    confirm_verify = get_config(swimmer.scm, C_SWIMMERS, C_CONF_DIFF, C_VERIFY)
    max_age = get_config(swimmer.scm, C_SWIMMERS, C_PARENT, C_MAX_AGE)

    if swimmer.email:
        email = swimmer.email.split(";")

    for parent in swimmer.parents:
        count += 1
        if parent.is_active is False:
            issue(parent, E_INACTIVE, f"Swimmer {swimmer.name}")

        if match is False:
            match = check_parent_email_match(email, parent)

        if confirm_verify:
            confirm_error = check_confirmed_diff(swimmer, parent)
            if confirm_error:
                issue(swimmer, E_CONFIRM_DIFF, f"Parent {parent.name}")
    if (
        swimmer.parents
        and swimmer.age
        and (match is False)
        and (swimmer.age <= max_age)
    ):
        if swimmer.print_exception(EXCEPTION_EMAILDIFF):
            err = f"{swimmer.email} - {swimmer.parents[0].email}"
            issue(swimmer, E_EMAIL_MATCH, err)

    if count == 0:
        mandatory = get_config(swimmer.scm, C_SWIMMERS, C_PARENT, C_MANDATORY)
        if mandatory and max_age:
            if swimmer.age and (swimmer.age <= max_age):
                msg = f"{swimmer.first_group}, Age: {swimmer.age}"
                issue(swimmer, E_NO_PARENT, msg)

    if count > 2:
        issue(swimmer, E_NUM_PARENTS)


def check_parent_email_match(email, parent):
    """Check swimmer and parent have email in common."""
    if parent.email:
        pemail = parent.email.split(";")
    else:
        return True  # can't match if non existent
    if email:
        for loop in email:
            for ploop in pemail:
                if loop == ploop:
                    return True
    return False


def check_confirmed_diff(swimmer, parent):
    """Check for differences in swimmer and parent."""
    # pylint: disable=R0911
    # Need them all
    child_mon = 0
    parent_mon = 0

    if swimmer.confirmed_date:
        child_mon = int((swimmer.confirmed_date.month - 1) / 3) * 3
    if parent.confirmed_date:
        parent_mon = int((parent.confirmed_date.month - 1) / 3) * 3

    if swimmer.age > get_config(swimmer.scm, C_SWIMMERS, C_PARENT, C_MAX_AGE):
        return False

    if child_mon == parent_mon:
        return False

    prefix = "Different confirmed dates"
    postfix = "- checking other details for consistency"
    debug(f"{prefix} {swimmer.name}, {parent.name} {postfix}", 8)

    if swimmer.email != parent.email:
        debug(f"email: {swimmer.email}: {parent.email}", 8)
        return True

    if swimmer.homephone != parent.homephone:
        debug(f"phone: {swimmer.homephone}: {parent.homephone}", 8)
        return True

    if swimmer.mobilephone != parent.mobilephone:
        debug(f"mobile: {swimmer.mobilephone}: {parent.mobilephone}", 8)
        return True

    if swimmer.address != parent.address:
        debug(f"address: {swimmer.address}: {parent.address}", 8)
        return True

    # Dates are different, but core attributes same
    # So set the confirmed date - to inhibit a confirm notice to parent

    if parent.confirmed_date is None:
        parent.set_confirmed(swimmer.confirmed_date)
        return False

    if swimmer.confirmed_date:
        if swimmer.confirmed_date > parent.confirmed_date:
            parent.set_confirmed(swimmer.confirmed_date)

    return False



"""Coach routines."""
from scm_helper.config import (
    A_ISCOACH,
    C_COACHES,
    C_MANDATORY,
    C_ROLE,
    EXCEPTION_NOSESSIONS,
    EXCEPTION_PERMISSIONS,
    get_config,
)
from scm_helper.issue import (
    E_COACH_WITH_SESSIONS,
    E_NO_ROLE_COACH,
    E_NO_SESSIONS,
    E_NOT_A_COACH,
    E_PERMISSION_EXTRA,
    E_PERMISSION_MISSING,
    debug,
    issue,
)


def analyse_coach(coach):
    """Analyse a coach..."""
    if coach.coach_role is False:
        if get_config(coach.scm, C_COACHES, C_ROLE, C_MANDATORY):
            issue(coach, E_NO_ROLE_COACH)

    if len(coach.coach_sessions) == 0:
        if coach.print_exception(EXCEPTION_NOSESSIONS):
            issue(coach, E_NO_SESSIONS)

    coach.check_dbs("Coach")


def check_coach_permissions(coach, role):
    """Check a coaches permissions."""
    debug(f"Permission check: {coach.name}, {role.name}", 7)
    if coach.is_coach is False:
        issue(coach, E_NOT_A_COACH, f"Role: {role.name} (fixable)")
        fix = {}
        fix[A_ISCOACH] = "1"
        coach.fixit(fix, "Add 'Is a coach'")

    coach.set_in_coach_role()

    if coach.is_swimmer is False:
        if len(coach.sessions) > 0:
            issue(coach, E_COACH_WITH_SESSIONS, f"Role: {role.name}")

    if coach.print_exception(EXCEPTION_PERMISSIONS) is False:
        return

    for session in coach.coach_sessions:
        match = False
        for permission in coach.restricted:
            if session == permission:
                match = True
                break
        if match is False:
            # SCM bug 6588
            issue(coach, E_PERMISSION_MISSING, session.full_name)
            # TODO
            # fix = {}
            # data = coach.data["SessionRestrictions"]
            # fix["SessionRestrictions"] = data.copy()
            # fix["SessionRestrictions"].append({A_GUID: session.guid})
            # coach.fixit(fix, f"Add permission for {session.name}")

    for permission in coach.restricted:
        match = False
        for session in coach.coach_sessions:
            if session == permission:
                match = True
                break
        if match is False:
            # SCM bug 6588
            issue(coach, E_PERMISSION_EXTRA, permission.full_name)
            # TODO
            # fix = {}
            # data = coach.data["SessionRestrictions"]
            # debug(f"Session restriction deletion - before:\n{data}\n", 9)
            # fix["SessionRestrictions"] = data.copy()
            # fix["SessionRestrictions"].remove({A_GUID: permission.guid})
            # coach.fixit(fix, f"Remove permission for {permission.name}")



"""SCM Group."""
import datetime

from scm_helper.config import (
    C_CHECK_DBS,
    C_CONFIRMATION,
    C_GROUP,
    C_GROUPS,
    C_IGNORE_GROUP,
    C_IGNORE_SWIMMER,
    C_MAX_AGE,
    C_MIN_AGE,
    C_NO_CLUB_SESSIONS,
    C_NO_SESSION_ALLOWED,
    C_NO_SESSIONS,
    C_SESSIONS,
    C_TYPE,
    CTYPE_COACH,
    CTYPE_SWIMMER,
    EXCEPTION_GROUPNOSESSION,
    EXCEPTION_NONSWIMMINGMASTER,
    SCM_CSV_DATE_FORMAT,
    get_config,
)
from scm_helper.entity import Entities, Entity, check_type
from scm_helper.issue import (
    E_CONFIRMATION_EXPIRED,
    E_NO_SWIMMERS,
    E_NOT_IN_SESSION,
    E_SESSIONS,
    E_TYPE,
    debug,
    debug_trace,
    issue,
)
from scm_helper.notify import notify

A_GROUP_NAME = "GroupName"


class Groups(Entities):
    """Groups."""

    def new_entity(self, entity):
        """Create a new entity."""
        return Group(entity, self.scm, self._url)


class Group(Entity):
    """A group."""

    def linkage(self, members):
        """Link members."""
        super().linkage(members)

        no_session = self.config_item(C_NO_CLUB_SESSIONS)
        ignore_group = self.config_item(C_IGNORE_GROUP)
        ignore_swimmer = self.config_item(C_IGNORE_SWIMMER)

        if self.members:
            for swimmer in self.members:
                swimmer.add_group(self)
                if no_session:
                    swimmer.set_no_session_ok()
                if ignore_swimmer:
                    swimmer.set_ignore_swimmer(True)
                if ignore_group:
                    swimmer.set_ignore_group(True)

    @debug_trace(5)
    def analyse(self):
        """Analyse the group."""
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-statements
        # pylint: disable=too-many-locals
        no_session = False
        check_dbs = False
        wanted_session = None
        allowed = None
        xtype = None
        ignore = None
        confirm = None

        if self.config:
            ignore = self.config_item(C_IGNORE_GROUP)
            no_session = self.config_item(C_NO_SESSIONS)
            check_dbs = self.config_item(C_CHECK_DBS)
            wanted_sessions = self.config_item(C_SESSIONS)
            allowed = self.config_item(C_NO_SESSION_ALLOWED)
            xtype = self.config_item(C_TYPE)
            confirm = self.config_item(C_CONFIRMATION)

        if ignore:
            debug(f"Ignoring group {self.name}", 7)
            return

        if len(self.members) == 0:
            issue(self, E_NO_SWIMMERS, "Group")
            return

        if no_session:
            for member in self.members:
                if len(member.sessions) > 0:
                    name = member.sessions[0].name
                    issue(member, E_SESSIONS, f"Group: {self.name}, Session: {name}")

        if confirm:
            try:
                date = datetime.datetime.strptime(confirm, SCM_CSV_DATE_FORMAT)
                confirm = date
            except ValueError:
                notify(
                    f"*** Error in date format in config file for groups config: {confirm} ***\n"
                )
                confirm = None

        for member in self.members:
            self.check_age(member)
            if check_dbs:
                member.check_dbs(self.name)

            if confirm:
                err = False
                if member.confirmed_date:
                    gap = (confirm - member.confirmed_date).days
                    if gap >= 0:
                        err = True
                else:
                    err = True

                if err:
                    issue(member, E_CONFIRMATION_EXPIRED, f"Group: {self.name}")
                    msg = f"Confirmation Expired for Group: {member.name}"
                    member.scm.lists.add(msg, member)

            if member.newstarter:
                continue

            if wanted_session:
                for session in wanted_sessions:
                    if check_in_session(member, session, allowed) is False:
                        res1 = self.print_exception(EXCEPTION_NONSWIMMINGMASTER)
                        res2 = self.print_exception(EXCEPTION_GROUPNOSESSION)
                        if res1 or res2:
                            issue(member, E_NOT_IN_SESSION, f"Group: {self.name}")
                    break

            if xtype:
                if check_type(member, xtype):
                    continue
                if xtype == CTYPE_SWIMMER:  # if swimmers wanted, allow it to be a coach
                    if check_type(member, CTYPE_COACH) is False:
                        msg = f"Group: {self.name}, Type required: {xtype}"
                        issue(member, E_TYPE, msg)

    def check_age(self, swimmer):
        """Check in right age group."""
        mymin = 3
        mymax = 100

        max_age = self.config_item(C_MAX_AGE)
        if max_age:
            mymax = max_age

        min_age = self.config_item(C_MIN_AGE)
        if min_age:
            mymin = min_age

        if max_age or min_age:
            swimmer.check_age(mymin, mymax, self.name)

    @property
    def name(self):
        """Guid."""
        return self.data[A_GROUP_NAME]

    @property
    def config(self):
        """Own config parameters..."""
        return get_config(self.scm, C_GROUPS, C_GROUP, self.name)

    def config_item(self, item):
        """Own config parameters..."""
        return get_config(self.scm, C_GROUPS, C_GROUP, self.name, item)


def check_in_session(swimmer, wanted_session, allowed):
    """Check swimmer in session."""
    in_session = False
    if swimmer.find_session_substr(wanted_session):
        in_session = True

    if allowed and in_session is False:
        for allow in allowed:
            if swimmer.find_group(allow):
                in_session = True

    if in_session is False:
        if swimmer.no_session_ok:
            return True
        if swimmer.is_coach:
            return True
        return False

    return True



#!/usr/bin/python3
"""SCM support tools."""
import getopt
import sys

from scm_helper.api import API
from scm_helper.config import HELPURL
from scm_helper.facebook import Facebook
from scm_helper.file import Csv
from scm_helper.issue import REPORTS, IssueHandler
from scm_helper.notify import notify, set_notify
from scm_helper.records import Records
from scm_helper.sendmail import send_email
from scm_helper.version import VERSION

USAGE = f"""
scm <options>

Where <options> are:
   --analyse = run analysis on archive date
   --archive <date> = which archive to use in restore
   --backup = backup
   --coaches = report of coaches per session
   --confirm_email = print email addresses for confirm errors
   --covid = print a list of sessions with who has replied to the Covid declaration
   --csv <csvfile> = read and validate file
   --dump <type> = dump entities of <type>
   --error = print sorted by error
   --email = send report as an email
   --facebook = check Facebook membership issues
   -f --fix = fix issues (after confirmation of each)
   --format <format> = CSV or JSON dump format
   --help = help
   -l --lists = update lists
   -m, --member = print sorted by member
   --newstarter = report on new starters anyway (normally inhibited)
   --notes = print notes
   --password <password> = supply the password - useful for scripting.
   -q, --quiet = quiet mode
   --records = process records
   --newtimes <csvfile> = process new swim times into records
   --report <report> = which reports to run
   --restore <type> = restore an entity of <type> (need -archive as well)
   --se = Check against SE database
   --sessions = print a list of swimmers and their sessions in CSV format
   --to <email> = Who to send the emial to (used with --email)
   --verify <date> = use archive backup

SCM Helper Version: {VERSION}
For more help see: {HELPURL}
"""

SHORT_OPTS = "hlmfq"
LONG_OPTS = [
    "analyse",
    "archive=",
    "backup",
    "coaches",
    "confirm_email",
    "covid",
    "csv=",
    "dump=",
    "email",
    "error",
    "facebook",
    "fix",
    "format=",
    "help",
    "lists",
    "member",
    "newstarter",
    "newtimes=",
    "notes",
    "password=",
    "quiet",
    "records",
    "report=",
    "restore=",
    "se",
    "sessions",
    "to=",
    "verify=",
]

MAPPING = {
    "--archive": "--verify",
    "-h": "--help",
    "-q": "--quiet",
    "-m": "--member",
    "-l": "--lists",
    "-f": "--fix",
}


def parse_opts(argv, scm):
    """Parse Options."""
    try:
        opts, remainder = getopt.getopt(argv, SHORT_OPTS, LONG_OPTS)
    except getopt.GetoptError as error:

        print(f"Option error: {error}")
        print("Use --help for options")
        sys.exit(2)

    for opt, args in opts:
        opt = MAPPING.get(opt, opt)
        if opt == "--help":
            print(USAGE)
            sys.exit()
        else:
            scm.setopt(opt, args)

    if len(remainder) > 0:
        print(f"Unknown option: {remainder}")
        print("Use --help for options")
        sys.exit()


def cmd(argv=None):
    """Start everything."""
    # Yes, its complicated...
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-statements

    fbook = None
    csv = None

    if argv is None:
        argv = sys.argv[1:]

    issues = IssueHandler()

    # Initiate the API Class
    scm = API(issues)

    parse_opts(argv, scm)

    if scm.initialise(scm.option("--password")) is False:
        sys.exit()

    output = ""

    reports = scm.option("--report")
    if reports:
        if reports in REPORTS:
            reports = reports.lower()
        else:
            print(f"Unknown Report: {reports}")
            sys.exit(2)

    if scm.option("--backup"):
        if scm.backup_data():
            scm.print_summary(backup=True)
        sys.exit()

    if scm.option("--csv"):
        csv = Csv()
        if csv.readfile(scm.option("--csv"), scm) is False:
            sys.exit(2)

    if scm.option("--facebook"):
        fbook = Facebook()
        if fbook.read_data(scm) is False:
            sys.exit(2)

    if scm.option("--verify"):
        if scm.decrypt(scm.option("--verify")) is False:
            sys.exit(2)
    else:
        if scm.get_data(False) is False:
            sys.exit(2)

    if scm.option("--restore"):
        xtype = scm.option("--restore")
        if scm.restore(xtype):
            notify("Success.\n")
        sys.exit()

    if scm.option("--notes"):
        output = scm.members.print_notes()
        if scm.option("--email"):
            send_email(scm, output, "SCM: Notes")
        else:
            print(output)
        sys.exit()

    if scm.option("--dump"):
        what = scm.option("--dump")
        output = scm.dump(what)
        if scm.option("--email"):
            send_email(scm, output, f"SCM: Dump of {what}")
        else:
            print(output)
        sys.exit()

    quiet = False
    if scm.option("--quiet"):
        quiet = True
        set_notify(False)

    if scm.linkage() is False:
        sys.exit()

    if scm.option("--coaches"):
        output = scm.sessions.print_coaches()
        if scm.option("--email"):
            send_email(scm, output, "SCM: Coaches Report")
        else:
            print(output)
        sys.exit()

    if scm.option("--csv"):
        csv.analyse(scm)
        output = csv.print_errors()
        if scm.option("--email"):
            send_email(scm, output, "SCM: CSV Analysis")
        else:
            print(output)
        sys.exit()

    if scm.option("--se"):
        output = scm.se_check()
        if scm.option("--email"):
            send_email(scm, output, "SCM: SE Analysis")
        else:
            print(output)
        sys.exit()

    if scm.option("--facebook"):
        fbook.analyse()
        output = fbook.print_errors()
        if scm.option("--email"):
            send_email(scm, output, "SCM: Facebook Report")
        else:
            print(output)
        sys.exit()

    if scm.option("--records"):
        record = Records(scm)
        if record.read_baseline() is False:
            sys.exit(2)
        if scm.option("--newtimes"):
            if record.read_newtimes(scm.option("--newtimes")) is False:
                sys.exit(2)
        record.create_html()
        sys.exit()

    scm.analyse()

    if scm.option("--covid"):
        output = scm.sessions.print_swimmers_covid()
        if scm.option("--email"):
            send_email(scm, output, "SCM: Session / Covid Report")
        else:
            print(output)
        sys.exit()

    if scm.option("--sessions"):
        output = scm.members.print_swimmers_sessions()
        if scm.option("--email"):
            send_email(scm, output, "SCM: Swimmers Per Session Report")
        else:
            print(output)
        sys.exit()

    if scm.option("--lists"):
        if scm.option("--verify"):
            print("--lists can only be used with live data")
        else:
            scm.update()
        sys.exit()

    if scm.option("--confirm_email"):
        output = issues.confirm_email()
        if scm.option("--email"):
            send_email(scm, output, "SCM: Confirmation email addresses")
        else:
            print(output)
        sys.exit()

    if scm.option("--fix"):
        if scm.option("--verify"):
            print("--fix can only be used with live data")
        else:
            scm.apply_fixes()
        sys.exit()

    if scm.option("--errors"):
        output = issues.print_by_error(reports)
    elif scm.option("--member"):
        output = issues.print_by_name(reports)
    else:
        output = issues.print_by_error(reports)

    if scm.option("--email"):
        if reports:
            send_email(scm, output, f"SCM: {reports} report")
        else:
            send_email(scm, output, "SCM: Report")
    else:
        print(output)

    if quiet is False:
        print("Summary...")
        print(scm.print_summary())


def main():
    """Run scm-helper."""
    argv = sys.argv[1:]
    if len(argv) > 0:
        cmd()  # Command line options = run command line version
        sys.exit()

    try:
        # pylint: disable=import-outside-toplevel
        from tkinter import TclError, Tk
    except ImportError:
        cmd()
        sys.exit()

    try:
        root = Tk()
    except TclError:
        cmd()
        sys.exit()

    # pylint: disable=import-outside-toplevel
    from scm_helper.gui import ScmGui

    ScmGui(root)
    root.mainloop()


if __name__ == "__main__":
    main()



"""Parent routines."""
from scm_helper.config import (
    A_ISPARENT,
    A_USERNAME,
    C_AGE,
    C_CHILD,
    C_LOGIN,
    C_MANDATORY,
    C_MIN_AGE,
    C_PARENTS,
    get_config,
)
from scm_helper.issue import (
    E_INACTIVE,
    E_NO_CHILD,
    E_NO_LOGIN,
    E_PARENT_AGE,
    E_PARENT_AGE_TOO_OLD,
    issue,
)


def analyse_parent(parent):
    """Analyse a parent..."""
    # pylint: disable=too-many-branches
    active = False
    inactive = None

    for swimmer in parent.swimmers:
        if swimmer.is_active:
            active = True
        else:
            inactive = swimmer.name

    if active is False:
        if inactive is None:
            issue(parent, E_NO_CHILD, "fixable")
            fix = {}
            fix[A_ISPARENT] = "0"
            parent.fixit(fix, "Remove 'is parent'")
        else:
            issue(parent, E_INACTIVE, f"child {inactive}")

    newmember = True
    for swimmer in parent.swimmers:
        if swimmer.newstarter is False:
            newmember = False
            break

    if (newmember is True) and parent.swimmers:
        parent.set_joined_today()

    age = get_config(parent.scm, C_PARENTS, C_AGE, C_MIN_AGE)
    if parent.age and (parent.age < age):
        issue(parent, E_PARENT_AGE)

    age = get_config(parent.scm, C_PARENTS, C_AGE, C_CHILD)
    for swimmer in parent.swimmers:
        if active and swimmer.age and (swimmer.age >= age):
            issue(swimmer, E_PARENT_AGE_TOO_OLD, f"{swimmer.age}, {parent.name}")

    login = get_config(parent.scm, C_PARENTS, C_LOGIN, C_MANDATORY)
    if login and (parent.username is None) and parent.email:
        issue(parent, E_NO_LOGIN, "Parent (fixable)")
        fix = {}
        fix[A_USERNAME] = parent.email
        parent.fixit(fix, f"Create login, username: {parent.email}")



"""Notify."""
import sys

try:
    from tkinter import END, messagebox, simpledialog
except ImportError:
    pass  # ignore it, we'll be running command line only


WHERE = None

# Really horrid code, but a simple way of doing it.
# Can't import anything for this package either, otherwise an import loop.


def notify(msg):
    """Notify on STDERR."""
    # pylint: disable=global-statement
    global WHERE
    if WHERE:
        WHERE.insert(END, msg)
        WHERE.see(END)
    elif WHERE is None:
        sys.stderr.write(msg)
        sys.stderr.flush()
    # Where is False
    # Do nothing ==> quite mode.


def set_notify(where):
    """Where to report errors."""
    # pylint: disable=global-statement
    global WHERE
    WHERE = where


def interact(msg):
    """Get user input."""
    # pylint: disable=global-statement
    global WHERE
    if WHERE:
        prefix = "SCM-Helper: input needed"
        return simpledialog.askstring(prefix, msg, parent=WHERE.master)
    print(msg, end="")
    return input()


def interact_yesno(msg):
    """Get user input."""
    # pylint: disable=global-statement
    global WHERE
    if WHERE:
        msg += "?"
        return messagebox.askyesno("SCM-Helper: Yes / No?", msg, parent=WHERE.master)
    msg += " (y/n)?"
    print(msg, end="")

    resp = input()
    if resp == "y":
        return True
    return False



"""SCM Session."""
import datetime

from scm_helper.config import (
    A_ARCHIVED,
    A_DATEAGREED,
    A_GUID,
    A_LAST_ATTENDED,
    A_MAX_MEMBERS,
    A_MEMBERS,
    C_ABSENCE,
    C_CONDUCT,
    C_COVID,
    C_DATE,
    C_GROUPS,
    C_IGNORE_ATTENDANCE,
    C_REGISTER,
    C_SESSION,
    C_SESSIONS,
    PRINT_DATE_FORMAT,
    SCM_DATE_FORMAT,
    get_config,
)
from scm_helper.entity import Entities, Entity
from scm_helper.issue import (
    E_INACTIVE,
    E_NEVER_ATTENDED,
    E_NO_COACH,
    E_NO_REGISTER,
    E_NO_SWIMMERS,
    E_NOT_ATTENDED,
    E_NOT_IN_GROUP,
    E_TOO_MANY_SWIMMERS,
    debug_trace,
    issue,
)
from scm_helper.notify import notify

A_SESSION_NAME = "SessionName"
A_COACHES = "Coaches"


class Sessions(Entities):
    """Sessions."""

    def new_entity(self, entity):
        """Create a new entity."""
        return Session(entity, self.scm, self._url)

    def find_session_substr(self, substr):
        """Find a session from a substring."""
        for session in self.entities:
            if session.name.find(substr) >= 0:
                return session
        return None

    def print_coaches(self):
        """Print coaches per session."""
        res = ""
        for session in sorted(self.entities, key=lambda x: x.full_name):
            if session.is_active:
                res += f"{session.full_name}:\n"
                res += session.print_coaches()
                res += "\n"

        return res

    def print_swimmers_covid(self):
        """Print coaches per session."""
        covid = get_config(self.scm, C_SESSIONS, C_COVID)
        if covid is None:
            notify("Missing config for COVID option")
            return ""

        res = ""
        for session in sorted(self.entities, key=lambda x: x.full_name):
            if session.is_active:
                res += f"{session.full_name}\n"
                res += session.print_swimmer_covid()
                res += "\n"

        return res


class Session(Entity):
    """A session."""

    def __init__(self, entity, scm, url):
        """Initialise."""
        super().__init__(entity, scm, url)
        self.coaches = []
        self.swimmers = []

        self.ignore_attendance = False

        cfg = get_config(
            self.scm, C_SESSIONS, C_SESSION, self.name, C_IGNORE_ATTENDANCE
        )
        if cfg:
            self.ignore_attendance = cfg

    def linkage(self, members):
        """Link coaches and swimmers."""
        if self.is_active is False:
            return

        super().linkage(members)

        if self.data[A_COACHES]:
            for coach in self.data[A_COACHES]:
                guid = members.by_guid[coach[A_GUID]]
                if guid.is_active:
                    self.coaches.append(guid)
                    guid.add_coach_session(self)
                else:
                    if self.ignore_attendance is False:
                        issue(self, E_INACTIVE, "(Coach)")
                if A_LAST_ATTENDED in coach:
                    lastseen = coach[A_LAST_ATTENDED]
                    if lastseen:
                        guid.set_lastseen(lastseen)
        else:
            if self.is_active:
                issue(self, E_NO_COACH)

        # super() has already done member links, now search for last seen...
        if A_MEMBERS in self.data:
            for swimmer in self.data[A_MEMBERS]:
                guid = members.by_guid[swimmer[A_GUID]]
                guid.add_session(self)
                if A_LAST_ATTENDED in swimmer:
                    lastseen = swimmer[A_LAST_ATTENDED]
                    if lastseen:
                        guid.set_lastseen(lastseen)

    def print_coaches(self):
        """Print coaches."""
        res = ""
        absence = get_config(self.scm, C_SESSIONS, C_ABSENCE)
        for coach in self.data[A_COACHES]:
            guid = self.scm.members.by_guid[coach[A_GUID]]
            if guid.is_active:
                msg = ""
                lastseen = coach.get(A_LAST_ATTENDED, None)
                if lastseen:
                    when = datetime.datetime.strptime(lastseen, SCM_DATE_FORMAT)
                    if (self.scm.today - when).days > absence:
                        msg = f"(Lastseen: {lastseen})"
                else:
                    msg = "(Never seen)"
                res += f"   {guid.name} {msg}\n"

        return res

    def print_swimmer_covid(self):
        """Print swimmers."""
        # pylint: disable=too-many-nested-blocks
        # pylint: disable=too-many-branches
        res = ""
        covid = get_config(self.scm, C_SESSIONS, C_COVID)
        c_date_str = get_config(self.scm, C_CONDUCT, covid, C_DATE)
        if c_date_str is None:
            c_date_str = "1900-01-01"
        c_date = datetime.datetime.strptime(c_date_str, SCM_DATE_FORMAT)

        res += "  Coaches:\n"
        for coach in self.data[A_COACHES]:
            swimmer = self.scm.members.by_guid[coach[A_GUID]]
            if swimmer.is_active:
                msg = "No"
                code = swimmer.get_conduct_name(covid)
                if code:
                    code_members = code.data[A_MEMBERS]
                    for member in code_members:
                        code_member = self.scm.members.by_guid[member[A_GUID]]
                        if code_member == swimmer:
                            if member[A_DATEAGREED]:
                                m_date = datetime.datetime.strptime(
                                    member[A_DATEAGREED], SCM_DATE_FORMAT
                                )
                                if m_date > c_date:
                                    msg = "Yes"

                res += f"   {swimmer.name}, {msg}\n"

        res += "  Swimmers:\n"
        for swimmer in self.members:
            if swimmer.is_active:
                msg = "No"
                code = swimmer.get_conduct_name(covid)
                if code:
                    code_members = code.data[A_MEMBERS]
                    for member in code_members:
                        code_member = self.scm.members.by_guid[member[A_GUID]]
                        if code_member == swimmer:
                            if member[A_DATEAGREED]:
                                m_date = datetime.datetime.strptime(
                                    member[A_DATEAGREED], SCM_DATE_FORMAT
                                )
                                if m_date > c_date:
                                    msg = "Yes"

                res += f"   {swimmer.name}, {msg}\n"

        return res

    @debug_trace(5)
    def analyse(self):
        """Analyse the session."""
        # pylint: disable=too-many-branches
        if self.is_active is False:
            return

        if len(self.members) == 0:
            issue(self, E_NO_SWIMMERS, "Session")
            return

        cfg = get_config(self.scm, C_SESSIONS)
        absence = cfg.get(C_ABSENCE, 9999)
        register = cfg.get(C_REGISTER, 9999)

        groups = get_config(self.scm, C_SESSIONS, C_SESSION, self.name, C_GROUPS)

        seen = None
        n_swimmers = 0

        for swimmer in self.data[A_MEMBERS]:
            found = False
            lastseen = None
            person = self.scm.members.by_guid[swimmer[A_GUID]]

            n_swimmers += 1

            if person.newstarter:
                continue

            if groups:
                for group in groups:
                    if person.find_group(group):
                        found = True
                        break

                if found is False:
                    issue(
                        person,
                        E_NOT_IN_GROUP,
                        f"{self.full_name}",
                        0,
                        person.print_groups,
                    )

            if self.ignore_attendance is False:
                attr = swimmer.get(A_LAST_ATTENDED, False)
                if attr:
                    lastseen = datetime.datetime.strptime(attr, SCM_DATE_FORMAT)
                    if (self.scm.today - lastseen).days > absence:
                        msg = lastseen.strftime(PRINT_DATE_FORMAT)
                        issue(
                            person,
                            E_NOT_ATTENDED,
                            f"{self.full_name}",
                            0,
                            f"Last seen: {msg}",
                        )
                    if (seen is None) or (lastseen > seen):
                        seen = lastseen
                else:
                    if absence != 9999:
                        issue(person, E_NEVER_ATTENDED, f"{self.full_name}")

        if n_swimmers > self.max_members:
            issue(self, E_TOO_MANY_SWIMMERS, f"{n_swimmers} > {self.max_members}")

        if (self.ignore_attendance is False) and (register != 9999):
            msg = "Never"
            if seen:
                if (self.scm.today - seen).days > register:
                    msg = seen.strftime(PRINT_DATE_FORMAT)
                else:
                    return
            issue(self, E_NO_REGISTER, f"last taken: {msg} ({self.full_name})")

    @property
    def name(self):
        """Name."""
        return self.data[A_SESSION_NAME]

    @property
    def full_name(self):
        """Full session name."""
        name = self.data[A_SESSION_NAME]
        weekday = self.data["WeekDay"]
        location = self.data["SessionLocation"]
        time = self.data["StartTime"]
        return f"{name}, {weekday}, {location}, {time}"

    @property
    def is_active(self):
        """Is the entry active..."""
        if A_ARCHIVED in self.data:
            if self.data[A_ARCHIVED] == 0:
                return True
        return False

    @property
    def max_members(self):
        """Is the entry active..."""
        if A_MAX_MEMBERS in self.data:
            return self.data[A_MAX_MEMBERS]

        return 9999



"""Interface to SCM API."""
import os.path
import platform
from datetime import date, datetime
from pathlib import Path
from shutil import copyfile

import requests
import yaml

from scm_helper.conduct import CodesOfConduct
from scm_helper.config import (
    BACKUP_DIR,
    BACKUP_URLS,
    C_ALLOW_UPDATE,
    C_CLUB,
    C_DEBUG_LEVEL,
    CODES_OF_CONDUCT,
    CONFIG_DIR,
    CONFIG_FILE,
    GROUPS,
    KEYFILE,
    LISTS,
    MEMBERS,
    O_FIX,
    O_FORMAT,
    O_VERIFY,
    ROLES,
    SESSIONS,
    URL_CONDUCT,
    URL_GROUPS,
    URL_LISTS,
    URL_MEMBERS,
    URL_ROLES,
    URL_SESSIONS,
    URL_WHO,
    USER_AGENT,
    WHO,
    get_config,
    verify_schema,
    verify_schema_data,
)
from scm_helper.default import create_default_config
from scm_helper.entity import Entities, Who
from scm_helper.groups import Groups
from scm_helper.issue import debug, set_debug_level
from scm_helper.lists import Lists
from scm_helper.members import Members
from scm_helper.notify import notify
from scm_helper.roles import Roles
from scm_helper.sessions import Sessions
from scm_helper.version import VERSION


class API:
    """Main SCM object."""

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-public-methods
    # Need them all!

    def __init__(self, issues):
        """Initialize SCM main class."""
        # pylint: disable=too-many-instance-attributes
        # Need them all!
        self._options = {}
        self._config = None
        self._key = None
        self.groups = None
        self.lists = None
        self.roles = None
        self.sessions = None
        self.members = None
        self.conduct = None
        self.classes = []
        self.backup_classes = []
        self.class_byname = {}
        self.issue_handler = issues
        self.fixable = []
        self.crypto = None
        self.ipad = False

        self.today = datetime.now()
        q_month = (int((self.today.month - 1) / 3) * 3) + 1
        q_year = self.today.year
        self.eoy = datetime(int(q_year), 12, 31)
        offset = datetime(int(q_year), int(q_month), 1)
        self.q_offset = (self.today - offset).days

        if "iPad" in platform.machine():
            self.ipad = True

    def get_config_file(self):
        """Read configuration file."""
        home = str(Path.home())
        cfg = os.path.join(home, CONFIG_DIR, CONFIG_FILE)

        if os.path.isfile(cfg) is False:
            if create_default_config() is False:
                return False
            nmsg = "You will now be asked to provide a password.\n"
            nmsg += "This is used to protect the API key.\n"
            notify(nmsg)

        try:
            with open(cfg) as file:
                self._config = None
                self._config = yaml.safe_load(file)
        except EnvironmentError:
            notify(f"Cannot open configuration file: {cfg}\n")
            return False
        except (yaml.scanner.ScannerError, yaml.parser.ParserError) as error:
            notify(f"Error in configuration file: {error}\n")
            return False

        if verify_schema(self._config) is False:
            return False

        return True

    def get_config(self, password):
        """Get API key."""
        # pylint: disable=import-outside-toplevel
        if self._config is None:
            if self.get_config_file() is False:
                return False

        if self.ipad:
            from scm_helper.ipad import Crypto

            self.crypto = Crypto(self._config[C_CLUB], password)  # Salt
        else:
            from scm_helper.crypto import Crypto

            self.crypto = Crypto(self._config[C_CLUB], password)  # Salt

        home = str(Path.home())

        keyfile = os.path.join(home, CONFIG_DIR, KEYFILE)
        self._key = self.crypto.read_key(keyfile)
        if self._key is None:
            return False

        debug_config = self.config(C_DEBUG_LEVEL)
        set_debug_level(debug_config)

        debug(f"Quarter offset: {self.q_offset}", 9)

        return True

    def initialise(self, password):
        """Initialise."""

        if self.ipad:
            password = "dummy"  # Can't to crypto on iPad

        if self.get_config(password) is False:
            return False

        mapping = [
            [SESSIONS, URL_SESSIONS, Sessions],
            [GROUPS, URL_GROUPS, Groups],
            [LISTS, URL_LISTS, Lists],
            [ROLES, URL_ROLES, Roles],
            [CODES_OF_CONDUCT, URL_CONDUCT, CodesOfConduct],
            [MEMBERS, URL_MEMBERS, Members],
        ]

        for item in mapping:
            name, url, xclass = item
            res = xclass(self, name, url)
            self.classes.append(res)

            # Ugly, but can's see how else to do it
            if name == SESSIONS:
                self.sessions = res
            elif name == GROUPS:
                self.groups = res
            elif name == LISTS:
                self.lists = res
            elif name == ROLES:
                self.roles = res
            elif name == CODES_OF_CONDUCT:
                self.conduct = res
            elif name == MEMBERS:
                self.members = res

            name = name.rstrip("s")  # remove any plural!
            name = name.lower()
            self.class_byname[name] = res

        for xclass in BACKUP_URLS:
            name, url = xclass
            entity = Entities(self, name, url)
            self.backup_classes.append(entity)
            name = name.rstrip("s")  # remove any plural!
            name = name.lower()
            self.class_byname[name] = entity

        # Finally who's who
        entity = Who(self, WHO, URL_WHO)
        self.backup_classes.append(entity)
        name = WHO.lower()
        self.class_byname[name] = entity

        return True

    def get_data(self, backup):
        """Get data."""
        debug(f"(version: {VERSION})", 1)
        notify("Reading Data...\n")

        loop = self.classes
        if backup:
            loop = self.classes + self.backup_classes

        for aclass in loop:
            if aclass.get_data() is False:
                return False

        return True

    def get_members_only(self):
        """Get member data."""
        self.members.get_data()

    def se_check(self):
        """Get member data."""
        return self.members.se_check()

    def linkage(self):
        """Set up cross reference links between Entities."""
        notify("Linking...\n")

        for aclass in self.classes:
            aclass.linkage()

        if verify_schema_data(self) is False:
            return False
        return True

    def analyse(self):
        """Analyse the data."""
        notify("Analysing...\n")

        for aclass in self.classes:
            aclass.analyse()

        notify("Done.\n")

    def update(self):
        """Update (lists)."""
        notify("Updating...\n")
        self.lists.update()
        notify("Done.\n")

    def restore(self, xclass):
        """Restore data..."""
        if self.ipad:
            notify("Not implemented on iPad")
            return False

        xclass = xclass.lower()
        if xclass in self.class_byname:
            item = self.class_byname[xclass]
            notify(f"Restoring {item.name}...\n")
            return item.restore(xclass)
        notify(f"Backup type {xclass} not found\n")
        return False

    def dump(self, xclass):
        """Dump data..."""
        if self.ipad:
            notify("Not implemented on iPad")
            return False

        f_csv = "CSV"
        f_json = "JSON"
        xlist = [f_json, f_csv]
        xformat = f_json
        if self.option(O_FORMAT):
            xformat = self.option(O_FORMAT)

        if xformat not in xlist:
            notify(f"Unknown class {xformat}")

        xclass = xclass.lower()
        if xclass in self.class_byname:
            if xformat == f_json:
                return self.class_byname[xclass].pretty_print()
            return self.class_byname[xclass].csv()
        notify(f"Dump type {xclass} not found\n")
        return ""

    def delete(self):
        """Delete all entities."""
        delete = self.classes + self.backup_classes
        for aclass in delete:
            aclass.delete()

        self.issue_handler.delete()

        self.groups = None
        self.lists = None
        self.roles = None
        self.sessions = None
        self.members = None
        self.conduct = None
        self.classes = []
        self.backup_classes = []
        self.class_byname = {}
        self.fixable = []

    def backup_data(self):
        """Backup."""
        if self.ipad:
            notify("Not implemented on iPad")
            return False

        if self.get_data(True) is False:
            return False

        backup = self.classes + self.backup_classes
        for aclass in backup:
            if self.crypto.encrypt_backup(aclass.name, aclass.json) is False:
                return False

        # Backup config file too.
        home = str(Path.home())
        today = date.today()
        cfg = os.path.join(home, CONFIG_DIR)
        directory = os.path.join(home, CONFIG_DIR, BACKUP_DIR, f"{today}")

        src = os.path.join(cfg, CONFIG_FILE)
        dst = os.path.join(directory, CONFIG_FILE)
        copyfile(src, dst)

        # Backup keyfile file too.
        src = os.path.join(cfg, KEYFILE)
        dst = os.path.join(directory, KEYFILE)
        copyfile(src, dst)

        return True

    def decrypt(self, xdate):
        """Decrypt file."""
        if self.ipad:
            notify("Not implemented on iPad")
            return False

        restore = self.classes + self.backup_classes

        for aclass in restore:
            decrypted = self.crypto.decrypt_backup(aclass.name, xdate)
            if decrypted is None:
                return False
            aclass.parse_data(decrypted)

        notify("\n")
        return True

    def print_summary(self, backup=False):
        """Print summary."""
        debug("Print summary called", 6)
        output = ""
        for aclass in self.classes:
            output += aclass.print_summary()
        output += f"   Not confirmed: {self.members.count_not_confirmed}\n"

        if backup and self.backup_classes:
            for aclass in self.backup_classes:
                output += aclass.print_summary()

        if self.option(O_FIX):  # fixed them!
            return output

        if self.option(O_VERIFY):
            return output  # fixable not available with backup data

        length = len(self.fixable)
        if length > 0:
            output += f"\n{length} fixable errors...\n"
            output += self.list_fixes()

        output += "\n"
        return output

    def setopt(self, opt, args):
        """Set options."""
        if args:
            self._options[opt] = args
        else:
            self._options[opt] = True

    def api_read(self, url, page):
        """Read URL page."""
        club = self._config[C_CLUB]
        user_agent = USER_AGENT.replace("###CLUB_NAME###", club)

        headers = {
            "User-Agent": user_agent,
            "Authorization-Token": self._key,
            "Page": str(page),
        }

        debug(f"URL:\n{url}", 9)
        debug(f"Headers:\n{headers}", 8)

        response = requests.get(url, headers=headers)
        if response.ok:
            return response.json()

        if response.status_code == 404:  # Worked, but not found
            return False

        notify(f"\nErroring getting data from {url}, page:{page}\n")
        notify(response.reason)
        notify("\n")
        return None

    def api_write(self, entity, create):
        """Write data back to SCM."""
        club = self._config[C_CLUB]
        user_agent = USER_AGENT.replace("###CLUB_NAME###", club)

        headers = {
            "content-type": "application/json",
            "User-Agent": user_agent,
            "Authorization-Token": self._key,
        }

        if get_config(entity.scm, C_ALLOW_UPDATE) is False:
            notify("Update prohibited by config.\n")
            return None

        debug(f"URL:\n{entity.url}", 9)
        debug(f"Headers:\n{headers}", 8)

        data = entity.newdata
        if create:
            debug(f"Post request:\n{data}", 7)
            response = requests.post(entity.url, json=data, headers=headers)
        else:
            debug(f"Put request:\n{data}", 7)
            response = requests.put(entity.url, json=data, headers=headers)
        if response.ok:
            return response

        if response.status_code == 404:  # Worked, but not found
            return False

        notify(f"\nErroring posting data {entity.name}\n")
        notify(response.reason)
        notify("\n")
        return None

    def fix_search(self):
        """fix_search_index."""

        home = str(Path.home())
        cfg = os.path.join(home, CONFIG_DIR, "fixed_search.txt")
        if os.path.isfile(cfg) is True:
            notify("Not required - already fixed")
            return False

        res = self.members.fix_search()
        if res is False:
            return res

        with open(cfg, mode="w") as file:
            file.write(f"Fixed index: {self.today}")

        msg = "Index recreated - give SCM time to process changes before testing."
        notify(f"\n{msg}\n")

        return True

    def apply_fixes(self):
        """Apply any fixes."""
        if len(self.fixable) == 0:
            notify("Nothing to fix\n")
            return False

        for fix in self.fixable:
            if fix.apply_fix() is None:
                self.fixable = []
                return False

        self.fixable = []
        return True

    def list_fixes(self):
        """List any fixes."""
        if len(self.fixable) == 0:
            notify("Nothing to fix\n")
            return False

        res = ""
        for fix in self.fixable:
            res += f"{fix.name}: {fix.fixmsg}\n"

        return res

    def option(self, option):
        """Options."""
        if option in self._options:
            return self._options[option]
        return None

    def config(self, option):
        """Options."""
        if option in self._config:
            return self._config[option]
        return None



"""SCM entity superclass."""
import csv
import datetime
import json
import pprint
import sys

from scm_helper.config import (
    A_ACTIVE,
    A_ARCHIVED,
    A_GUID,
    A_MEMBERS,
    CTYPE_COACH,
    CTYPE_COMMITTEE,
    CTYPE_LIFESAVER,
    CTYPE_LTS,
    CTYPE_OPENWATER,
    CTYPE_PARENT,
    CTYPE_POLO,
    CTYPE_SWIMMER,
    CTYPE_SYNCHRO,
    CTYPE_VOLUNTEER,
    SCM_DATE_FORMAT,
)
from scm_helper.issue import E_INACTIVE, debug, debug_trace, issue
from scm_helper.notify import interact, interact_yesno, notify


class Entities:
    """Superclass for Entities."""

    # pylint: disable=too-many-instance-attributes
    # Need them all!

    def __init__(self, scm, name, url):
        """Initialize."""
        self.entities = []
        self.by_guid = {}
        self.by_name = {}
        self.knownas = {}
        self.scm = scm
        self._name = name
        self._url = url
        self.count = 0
        self._raw_data = []

    def get_data(self):
        """Get data."""
        notify(f"{self._name}... ")
        loop = 100
        page = 1
        count = 0
        while loop == 100:
            if count != 0:
                notify(f"{count} ")
            data = self.scm.api_read(self._url, page)
            if data is None:
                return False
            loop = self.create_entities(data)
            self._raw_data += data
            page += 1
            count += loop

        notify(f"{count}\n")
        return True

    def parse_data(self, data):
        """Read data."""
        notify(f"{self._name}...\n")
        self._raw_data = data
        self.create_entities(self._raw_data)

    def create_entities(self, entities):
        """Create entities."""
        i = 0
        for entity in entities:
            data = self.new_entity(entity)
            if data:
                self.entities.append(data)
                if data.guid:  # Who's who does not have a GUID
                    self.by_guid[data.guid] = data
                self.by_name[data.name] = data
                if data.is_active:
                    self.count += 1
            i += 1
        return i

    def new_entity(self, entity):
        """Create a new entity - OVERRIDE expected."""
        return Entity(entity, self.scm, self._url)

    def analyse(self):
        """Analyse the members."""
        for entity in self.entities:
            entity.analyse()

    def linkage(self):
        """Create Member links."""
        for entity in self.entities:
            entity.linkage(self.scm.members)

    def print_name(self):
        """Print name."""
        for entity in self.entities:
            print(entity.name)

    def delete(self):
        """Delete all members."""
        for entity in self.entities:
            entity.delete()
            del entity
        self.entities = []
        self.by_guid = {}
        self.by_name = {}
        self.knownas = {}
        self.count = 0
        self._raw_data = []

    def restore(self, xtype):
        """Restore a record."""
        name = interact(f"Enter name of {xtype} to restore: ")
        if name in self.by_name:
            return self.by_name[name].restore(self._url)
        if name in self.knownas:
            return self.knownas[name].restore(self._url)

        notify(f"{name} not found in backup of {xtype}.\n")
        return False

    @property
    def json(self):
        """JSON dump."""
        return json.dumps(self._raw_data)

    def pretty_print(self):
        """JSON dump."""
        print(json.dumps(self._raw_data, indent=4))

    def csv(self):
        """CSV dump."""
        output = csv.writer(sys.stdout)
        output.writerow(self._raw_data[0].keys())  # header row
        for row in self._raw_data:
            output.writerow(row.values())  # values row

    def print_summary(self):
        """Print a summary."""
        return f"{self._name}: {self.count}\n"

    @property
    def name(self):
        """Return name."""
        return self._name


class Entity:
    """A entity."""

    def __init__(self, entity, scm, url):
        """Initialize."""
        self.data = entity
        self.newdata = None  # used if updating
        self.fixmsg = ""
        self.url = f"{url}/{self.guid}"
        self.members = []
        self._scm = scm

    def print_exception(self, exception):
        """Is the exception allowable."""
        # pylint: disable=unused-argument
        # pylint: disable=no-self-use
        # Override uses exception
        return True

    @debug_trace(6)
    def linkage(self, members):
        """Link members."""
        if (A_MEMBERS in self.data) and (len(self.data[A_MEMBERS]) > 0):
            for swimmer in self.data[A_MEMBERS]:
                if swimmer[A_GUID] not in members.by_guid:
                    msg = (
                        f"GUID {swimmer[A_GUID]} missing in list - email address only?"
                    )
                    debug(msg, 7)
                    continue
                guid = members.by_guid[swimmer[A_GUID]]
                if guid.is_active:
                    self.members.append(guid)
                else:
                    name = guid.name
                    issue(self, E_INACTIVE, f"member {name}", 0, "Fixable")

                    if self.newdata and (A_MEMBERS in self.newdata):
                        fix = self.newdata
                    else:
                        fix = {}
                        fix[A_MEMBERS] = self.data[A_MEMBERS].copy()
                    remove = {A_GUID: guid.guid}
                    fix[A_MEMBERS].remove(remove)
                    self.fixit(fix, f"Delete {guid.name} (inactive)")

    def check_attribute(self, attribute):
        """Return the value, if there is one."""
        if attribute in self.data:
            if self.data[attribute]:
                return self.data[attribute]
        return None

    def set_date(self, field):
        """Set a date."""
        date = self.check_attribute(field)
        if date is None:
            return None
        return datetime.datetime.strptime(date, SCM_DATE_FORMAT)

    def delete(self):
        """Delete."""
        # pylint: disable=no-self-use
        # override if anything needs doing
        return

    def print_links(self):
        """Print links."""
        # pylint: disable=no-self-use
        # override if anything needs doing
        return ""

    def restore(self):
        """Update a record from backup."""
        # pylint: disable=too-many-return-statements
        self.newdata = self.data

        old = self.scm.api_read(self.url, 1)
        if old is None:
            notify("Error: Something went wrong connecting to SCM!\n")
            return False

        # API inconsistent in use of Archive or Active - check both.
        if A_ACTIVE in self.newdata:
            self.newdata[A_ACTIVE] = "1"  # Make sure active

        if A_ARCHIVED in self.newdata:
            self.newdata[A_ARCHIVED] = "1"  # Make sure active

        if old is False:
            del self.newdata[A_GUID]  # Will need a new GUID
            # SCM does not provide the PUSH/PUT options for sessions and groups
            # Would be a lot of code anyway - is it needed?
            # So print help...
            notify(self.print_links())
            return self.scm.api_write(self, True)

        if A_ACTIVE in old:
            if old[A_ACTIVE] == "1":
                return self.scm.api_write(self, False)

            notify("Error: Already exist, just inactive.  Re-activate in SCM.\n")
            return False

        if A_ARCHIVED in old:
            if old[A_ARCHIVED] == 0:
                # API is inconsistent here too - no quotes for this one.
                return self.scm.api_write(self, False)

            notify("Error: Already exist, just archived.  Re-activate in SCM\n")
            return False

        return self.scm.api_write(self, False)

    def fixit(self, fix, message):
        """Prepare to fix an entity."""
        if self.newdata is None:
            self.newdata = fix
            self.fixmsg = message
        else:
            self.newdata.update(fix)
            self.fixmsg += f", {message}"

        if self in self.scm.fixable:
            return
        self.scm.fixable.append(self)

    def apply_fix(self):
        """Fix an entity."""
        printer = pprint.PrettyPrinter(indent=4)
        data = printer.pformat(self.newdata)
        err = f"Fix '{self.name}' with:\n    {self.fixmsg}\nConfirm"
        debug("fixit:", 7)
        debug(data, 8)
        resp = interact_yesno(err)
        if resp is False:
            return False

        self.newdata[A_GUID] = self.guid

        notify(f"Fixing: {self.name}...")

        res = self.scm.api_write(self, False)

        if res:
            notify("Success.\n")
        return res

    @property
    def guid(self):
        """Guid."""
        if A_GUID in self.data:
            return self.data[A_GUID]
        return None

    @property
    def name(self):
        """Must override."""
        return "Name error - no override"

    @property
    def full_name(self):
        """Return full name (if overridden)."""
        return self.name

    @property
    def is_active(self):
        """Is the entity active - True unless overridden."""
        return True

    @property
    def newstarter(self):
        """Is the member a new stater - override in members.py."""
        return False

    @property
    def scm(self):
        """Return SCM entry."""
        return self._scm


# Not part of class
def check_type(member, xtype):
    """Check each type for a match."""
    # pylint: disable=too-many-return-statements
    # Need them all!
    if xtype == CTYPE_SWIMMER:
        return member.is_swimmer

    if xtype == CTYPE_SYNCHRO:
        return member.is_synchro

    if xtype == CTYPE_COACH:
        return member.is_coach

    if xtype == CTYPE_COMMITTEE:
        return member.is_committee

    if xtype == CTYPE_OPENWATER:
        return member.is_openwater

    if xtype == CTYPE_POLO:
        return member.is_polo

    if xtype == CTYPE_PARENT:
        return member.is_parent

    if xtype == CTYPE_VOLUNTEER:
        return member.is_volunteer

    if xtype == CTYPE_LTS:
        return member.is_lts

    if xtype == CTYPE_LIFESAVER:
        return member.is_lifesaver

    return False


def get_date(date, xformat):
    """Parse a date."""
    try:
        res = datetime.datetime.strptime(date, xformat)
        return res
    except ValueError as error:
        notify(f"Error in date: {error}")
        return None


def print_all(xlist):
    """Print out the entity."""
    grp = None
    for obj in xlist:
        if grp:
            grp += f", {obj.name}"
        else:
            grp = obj.name
    if grp is None:
        grp = "None"

    return grp


class Who(Entities):
    """Subclass for Who's who."""

    # pylint: disable=too-many-instance-attributes
    # Need them all!

    def get_data(self):
        """Get data."""
        notify(f"{self._name}... ")

        data = self.scm.api_read(self._url, 1)
        if data is None:
            return False
        count = self.create_entities(data)
        # line below is subtly different, who's who data is already a list.
        self._raw_data = data

        notify(f"{count}\n")
        if count != 1:
            debug("Who's who assumption failure", 0)

        return True



"""Issue handling."""
from datetime import datetime

from scm_helper.config import C_IGNORE_ERROR, C_ISSUES, EXCEPTION_GENERAL, O_NEWSTARTER
from scm_helper.notify import notify

R_COACH = "coaches"
R_CONFIRMATION = "confirmation"
R_DBS = "dbs"
R_GROUP = "groups"
R_MEMBER = "members"
R_SESSION = "sessions"
R_LIST = "lists"
R_ROLE = "roles"

REPORTS = [R_MEMBER, R_COACH, R_DBS, R_GROUP, R_SESSION, R_LIST, R_ROLE, R_CONFIRMATION]

R_PRINT = {
    R_COACH: "Coaches Report",
    R_CONFIRMATION: "Confirmation Report",
    R_DBS: "DBS Report",
    R_GROUP: "Groups Report",
    R_MEMBER: "Members Report",
    R_SESSION: "Sessions Report",
    R_LIST: "Lists Report",
    R_ROLE: "Roles Report",
}

NAME = "name"
MESSAGE = "message"
REPORT = "report"
REVERSE = "reverse"

# errors

# Black directive
# fmt: off
E_ABSENT = {
    NAME: "E_ABSENT",
    MESSAGE: "Absent",
    REVERSE: False,
    REPORT: R_MEMBER}
E_ASA = {
    NAME: "E_ASA",
    MESSAGE: "No Swim England number",
    REVERSE: False,
    REPORT: R_MEMBER}
E_COACH_WITH_SESSIONS = {
    NAME: "E_COACH_WITH_SESSIONS",
    MESSAGE: "Coach with swim sessions in profile",
    REVERSE: False,
    REPORT: R_COACH}
E_COACH_ROLE = {
    NAME: "E_COACH_ROLE",
    MESSAGE: "In coach role, but not a coach",
    REVERSE: False,
    REPORT: R_ROLE}
E_CONFIRM_DIFF = {
    NAME: "E_CONFIRM_DIFF",
    MESSAGE: "Difference in confirmation dates",
    REVERSE: False,
    REPORT: R_CONFIRMATION}
E_CONFIRMATION_EXPIRED = {
    NAME: "E_CONFIRMATION_EXPIRED",
    MESSAGE: "Confirmation expired",
    REVERSE: True,
    REPORT: R_CONFIRMATION}
E_DATE = {
    NAME: "E_DATE",
    MESSAGE: "Error in date format",
    REVERSE: False,
    REPORT: R_MEMBER}
E_DATE_JOINED = {
    NAME: "E_DATE_JOINED",
    MESSAGE: "No date joined",
    REVERSE: False,
    REPORT: R_MEMBER}
E_DBS_EXPIRED = {
    NAME: "E_DBS_EXPIRED",
    MESSAGE: "DBS Expiring/Expired",
    REVERSE: False,
    REPORT: R_DBS}
E_DOB = {
    NAME: "E_DOB",
    MESSAGE: "No date of birth",
    REVERSE: False,
    REPORT: R_MEMBER}
E_DUPLICATE = {
    NAME: "E_DUPLICATE",
    MESSAGE: "Duplicate user",
    REVERSE: False,
    REPORT: R_MEMBER}
E_EMAIL_MATCH = {
    NAME: "E_EMAIL_MATCH",
    MESSAGE: "Swimmers email does not match parents",
    REVERSE: False,
    REPORT: R_MEMBER}
E_EMAIL_SPACE = {
    NAME: "E_EMAIL_SPACE",
    MESSAGE: "Space in email",
    REVERSE: False,
    REPORT: R_MEMBER}
E_GENDER = {
    NAME: "E_GENDER",
    MESSAGE: "No Gender",
    REVERSE: False,
    REPORT: R_MEMBER}
E_INACTIVE = {
    NAME: "E_INACTIVE",
    MESSAGE: "Inactive",
    REVERSE: False,
    REPORT: R_MEMBER}
E_INACTIVE_TOOLONG = {
    NAME: "E_INACTIVE_TOOLONG",
    MESSAGE: "Inactive for too long - consider archiving",
    REVERSE: False,
    REPORT: R_MEMBER}
E_LIST_ERROR = {
    NAME: "E_LIST_ERROR",
    MESSAGE: "Error in email list",
    REVERSE: False,
    REPORT: R_LIST}
E_LOGIN_TOO_YOUNG = {
    NAME: "E_LOGIN_TOO_YOUNG",
    MESSAGE: "Swimmer with login: too young",
    REVERSE: False,
    REPORT: R_MEMBER}
E_JOB = {
    NAME: "E_JOB",
    MESSAGE: "Job description but not a volunteer",
    REVERSE: False,
    REPORT: R_MEMBER}
E_NAME_CAPITAL = {
    NAME: "E_NAME_CAPITAL",
    MESSAGE: "Check name capitilisation",
    REVERSE: False,
    REPORT: R_MEMBER}
E_MAX_SESSIONS = {
    NAME: "E_MAX_SESSIONS",
    MESSAGE: "Maximum nuber of sessions permitted exceeded",
    REVERSE: False,
    REPORT: R_SESSION}
E_NEVERSEEN = {
    NAME: "E_NEVERSEEN",
    MESSAGE: "Never seen",
    REVERSE: False,
    REPORT: R_MEMBER}
E_NEVER_ATTENDED = {
    NAME: "E_NEVER_ATTENDED",
    MESSAGE: "Never attended",
    REVERSE: True,
    REPORT: R_SESSION}
E_NO_CHILD = {
    NAME: "E_NO_CHILD",
    MESSAGE: "Parent, no children",
    REVERSE: False,
    REPORT: R_MEMBER}
E_NO_COACH = {
    NAME: "E_NO_COACH",
    MESSAGE: "No coach",
    REVERSE: False,
    REPORT: R_COACH}
E_NO_CONDUCT = {
    NAME: "E_NO_CONDUCT",
    MESSAGE: "Code of Conduct missing",
    REVERSE: True,
    REPORT: R_CONFIRMATION}
E_NO_CONDUCT_DATE = {
    NAME: "E_NO_CONDUCT_DATE",
    MESSAGE: "Code of Conduct not signed",
    REVERSE: True,
    REPORT: R_CONFIRMATION}
E_NO_DBS = {
    NAME: "E_NO_DBS",
    MESSAGE: "No DBS",
    REVERSE: False,
    REPORT: R_DBS}
E_NO_EMAIL = {
    NAME: "E_NO_EMAIL",
    MESSAGE: "No email address",
    REVERSE: False,
    REPORT: R_MEMBER}
E_NO_GROUP = {
    NAME: "E_NO_GROUP",
    MESSAGE: "Swimmer with no group",
    REVERSE: False,
    REPORT: R_GROUP}
E_NO_JOB = {
    NAME: "E_NO_JOB",
    MESSAGE: "No job description",
    REVERSE: False,
    REPORT: R_ROLE}
E_NO_LEAVE_DATE = {
    NAME: "E_NO_LEAVE_DATE",
    MESSAGE: "Inactive member with no leaving date",
    REVERSE: False,
    REPORT: R_MEMBER}
E_NO_LOGIN = {
    NAME: "E_NO_LOGIN",
    MESSAGE: "No login",
    REVERSE: False,
    REPORT: R_MEMBER}
E_NO_PARENT = {
    NAME: "E_NO_PARENT",
    MESSAGE: "Swimmer with no parent",
    REVERSE: False,
    REPORT: R_MEMBER}
E_NO_REGISTER = {
    NAME: "E_NO_REGISTER",
    MESSAGE: "Register not taken",
    REVERSE: False,
    REPORT: R_SESSION}
E_NO_RESTRICTIONS = {
    NAME: "E_NO_RESTRICTIONS",
    MESSAGE: "In role with no restrictions",
    REVERSE: False,
    REPORT: R_ROLE}
E_NO_ROLE_COACH = {
    NAME: "E_NO_ROLE_COACH",
    MESSAGE: "Coach not in a coach role",
    REVERSE: False,
    REPORT: R_ROLE}
E_NO_SAFEGUARD = {
    NAME: "E_NO_SAFEGUARD",
    MESSAGE: "No Safeguarding",
    REVERSE: False,
    REPORT: R_DBS}
E_NO_SESSIONS = {
    NAME: "E_NO_SESSIONS",
    MESSAGE: "Coach with no sessions",
    REVERSE: False,
    REPORT: R_COACH}
E_NO_SWIMMERS = {
    NAME: "E_NO_SWIMMERS",
    MESSAGE: "No swimmers",
    REVERSE: False,
    REPORT: R_SESSION}
E_TOO_MANY_SWIMMERS = {
    NAME: "E_TOO_MANY_SWIMMERS",
    MESSAGE: "Max swimmers exceeded",
    REVERSE: False,
    REPORT: R_SESSION}
E_NOT_A_COACH = {
    NAME: "E_NOT_A_COACH",
    MESSAGE: "Not a coach, but in coach role",
    REVERSE: False,
    REPORT: R_ROLE}
E_NOT_ATTENDED = {
    NAME: "E_NOT_ATTENDED",
    MESSAGE: "Swimmer not attended a session for given period of time",
    REVERSE: True,
    REPORT: R_SESSION}
E_NOT_CONFIRMED = {
    NAME: "E_NOT_CONFIRMED",
    MESSAGE: "Not confirmed",
    REVERSE: True,
    REPORT: R_CONFIRMATION}
E_NOT_IN_SESSION = {
    NAME: "E_NOT_IN_SESSION",
    MESSAGE: "Member not is any sessions for the group",
    REVERSE: True,
    REPORT: R_GROUP}
E_NOT_IN_GROUP = {
    NAME: "E_NOT_IN_GROUP",
    MESSAGE: "Member not in group required for the session",
    REVERSE: True,
    REPORT: R_GROUP}
E_TYPE_GROUP = {
    NAME: "E_TYPE_GROUP",
    MESSAGE: "Member not in group required for the type",
    REVERSE: True,
    REPORT: R_GROUP}
E_NUM_PARENTS = {
    NAME: "E_NUM_PARENTS",
    MESSAGE: "More than two parents",
    REVERSE: False,
    REPORT: R_MEMBER}
E_OWNPARENT = {
    NAME: "E_OWNPARENT",
    MESSAGE: "Swimmer is own parent",
    REVERSE: False,
    REPORT: R_MEMBER}
E_PARENT_AGE = {
    NAME: "E_PARENT_AGE",
    MESSAGE: "Parent too young",
    REVERSE: False,
    REPORT: R_MEMBER}
E_PARENT_AGE_TOO_OLD = {
    NAME: "E_PARENT_AGE_TOO_OLD",
    MESSAGE: "Child too old to have parent",
    REVERSE: False,
    REPORT: R_MEMBER}
E_PERMISSION_EXTRA = {
    NAME: "E_PERMISSION_EXTRA",
    MESSAGE: "Extra session permission",
    REVERSE: False,
    REPORT: R_ROLE}
E_PERMISSION_MISSING = {
    NAME: "E_PERMISSION_MISSING",
    MESSAGE: "Missing permission",
    REVERSE: False,
    REPORT: R_ROLE}
E_SAFEGUARD_EXPIRED = {
    NAME: "E_SAFEGUARD_EXPIRED",
    MESSAGE: "Safeguarding Expiring/Expired",
    REVERSE: False,
    REPORT: R_DBS}
E_SESSIONS = {
    NAME: "E_SESSIONS",
    MESSAGE: "Members has session, but in no sessions group",
    REVERSE: True,
    REPORT: R_SESSION}
E_TOO_OLD = {
    NAME: "E_TOO_OLD",
    MESSAGE: "Too old for group",
    REVERSE: False,
    REPORT: R_GROUP}
E_TOO_YOUNG = {
    NAME: "E_TOO_YOUNG",
    MESSAGE: "Too young for group",
    REVERSE: False,
    REPORT: R_GROUP}
E_TWO_GROUPS = {
    NAME: "E_TWO_GROUPS",
    MESSAGE: "Swimmers in two groups",
    REVERSE: True,
    REPORT: R_GROUP}
E_TYPE = {
    NAME: "E_TYPE",
    MESSAGE: "swimmer has wrong type for group",
    REVERSE: False,
    REPORT: R_GROUP}
E_TYPE_GROUP = {
    NAME: "E_TYPE_GROUP",
    MESSAGE: "Member not in group required for the type",
    REVERSE: True,
    REPORT: R_GROUP}
E_UNKNOWN = {
    NAME: "E_UNKNOWN",
    MESSAGE: "Not a swimmers/parent/coach/official.  Who are they?",
    REVERSE: False,
    REPORT: R_MEMBER}
E_UNUSED_LOGIN = {
    NAME: "E_UNUSED_LOGIN",
    MESSAGE: "Login not used",
    REVERSE: False,
    REPORT: R_ROLE}
E_VOLUNTEER = {
    NAME: "E_VOLUNTEER",
    MESSAGE: "Not marked as a volunteer",
    REVERSE: False,
    REPORT: R_ROLE}

# Black directive
# fmt: on

ISSUE_LIST = [
    E_ABSENT,
    E_ASA,
    E_COACH_ROLE,
    E_COACH_WITH_SESSIONS,
    E_CONFIRM_DIFF,
    E_CONFIRMATION_EXPIRED,
    E_DATE,
    E_DATE_JOINED,
    E_DBS_EXPIRED,
    E_DOB,
    E_DUPLICATE,
    E_EMAIL_MATCH,
    E_EMAIL_SPACE,
    E_GENDER,
    E_INACTIVE,
    E_INACTIVE_TOOLONG,
    E_JOB,
    E_LIST_ERROR,
    E_LOGIN_TOO_YOUNG,
    E_NAME_CAPITAL,
    E_NEVER_ATTENDED,
    E_NEVERSEEN,
    E_NO_CHILD,
    E_NO_COACH,
    E_NO_CONDUCT,
    E_NO_CONDUCT_DATE,
    E_NO_DBS,
    E_NO_EMAIL,
    E_NO_GROUP,
    E_NO_JOB,
    E_NO_LEAVE_DATE,
    E_NO_LOGIN,
    E_NO_PARENT,
    E_NO_REGISTER,
    E_NO_RESTRICTIONS,
    E_NO_ROLE_COACH,
    E_NO_SAFEGUARD,
    E_NO_SESSIONS,
    E_NO_SWIMMERS,
    E_NOT_A_COACH,
    E_NOT_ATTENDED,
    E_NOT_CONFIRMED,
    E_NOT_IN_GROUP,
    E_NOT_IN_SESSION,
    E_NUM_PARENTS,
    E_OWNPARENT,
    E_PARENT_AGE,
    E_PARENT_AGE_TOO_OLD,
    E_PERMISSION_EXTRA,
    E_PERMISSION_MISSING,
    E_SAFEGUARD_EXPIRED,
    E_SESSIONS,
    E_TOO_OLD,
    E_TOO_YOUNG,
    E_TWO_GROUPS,
    E_TYPE,
    E_TYPE_GROUP,
    E_TYPE_GROUP,
    E_UNKNOWN,
    E_UNUSED_LOGIN,
    E_VOLUNTEER,
]

# Handler
HANDLER = None


def issue(xobject, error, msg=None, level=0, msg2=""):
    """Record an issue."""

    debug(f"ISSUE: {xobject.name}, {error[MESSAGE]} / {msg}", 5)

    if level != -1:
        if xobject.print_exception(EXCEPTION_GENERAL) is False:
            debug(f"Error ignored due to exception {xobject.name}", 3)
            return

        if level > HANDLER.debug_level:
            return

        if xobject.newstarter:
            if xobject.scm.option(O_NEWSTARTER):
                pass
            else:
                prefix = "Error ignored - new starter - "
                debug(f"{prefix}{xobject.name}, {error[MESSAGE]} ({msg})", 3)
                return

    HANDLER.add_issue(xobject, error, msg, msg2)


def debug(msg, level):
    """Debug error handler."""
    if level > HANDLER.debug_level:
        return

    msg += "\n"
    notify(msg)


def set_debug_level(level):
    """Set debugging level."""
    if level is None:
        HANDLER.debug_level = 0
        return

    HANDLER.debug_level = level


def debug_trace(level):
    """Decorator to provide a trace capability."""

    def wrap(func):
        def wrapped_f(self, *args):
            name = func.__name__
            xclass = self.__class__.__name__
            xtime = datetime.now().time()
            debug(f"{xtime}: Entry {name}/{xclass}/{self.name}", level)
            retval = func(self, *args)
            debug(f"Exit {self.name}", level)
            return retval

        return wrapped_f

    return wrap


class IssueHandler:
    """report to handle issues."""

    def __init__(self):
        """Initialise."""
        self.issues = []
        self.by_name = {}
        self.by_error = {}
        self.debug_level = 0
        self._config = None
        self.scm = None

        global HANDLER  # pylint: disable=global-statement
        HANDLER = self

    def delete(self):
        """Clear database, ready for rerun."""
        self.issues = []
        self.by_name = {}
        self.by_error = {}

    def add_issue(self, xobject, error, msg, msg2):
        """Record an issue."""
        # Yes, its complicated...
        # pylint: disable=too-many-branches

        if self._config is None:
            self._config = xobject.scm.config(C_ISSUES)
        if self._config is None:
            self._config = []

        if self.scm is None:
            # pylint: disable=protected-access
            self.scm = xobject._scm  # Yuck, but otherwise a loop

        ignore = False

        if error[NAME] in self._config:
            if C_IGNORE_ERROR in self._config[error[NAME]]:
                ignore = self._config[error[NAME]][C_IGNORE_ERROR]
            if MESSAGE in self._config[error[NAME]]:
                error[MESSAGE] = self._config[error[NAME]][MESSAGE]

        if ignore:
            return

        err = error[MESSAGE]
        rpt = error[REPORT]
        rev = error[REVERSE]
        name = xobject.full_name
        create_dict(self.by_name, name, err, msg, msg2, rpt, False, xobject)
        if rev:
            create_dict(self.by_error, err, msg, name, msg2, rpt, rev, xobject)
        else:
            create_dict(self.by_error, err, name, msg, msg2, rpt, rev, xobject)

    def print_by_name(self, reports):
        """Print all issues by name."""
        debug(f"Print by name called {reports}", 6)
        return print_dict(self.by_name, reports)

    def print_by_error(self, reports):
        """Print all issues by error."""
        debug(f"Print by error called {reports}", 6)
        if reports is None:
            res = ""
            for report in REPORTS:
                res += f"========= {R_PRINT[report]} ========\n"
                res += print_dict(self.by_error, report)
            res += "======================="
            return res

        return print_dict(self.by_error, reports)

    def check_issue(self, xissue):
        """Check if this is a valid issue."""
        # pylint: disable=no-self-use
        # if it was a function, you end up with circualr imports
        for anissue in ISSUE_LIST:
            if anissue[NAME] == xissue:
                return True
        return False

    def confirm_email(self):
        """Print email addresses for confirmation errors."""
        matrix = {}
        for anissue in self.by_error:
            for anerror in self.by_error[anissue]:
                pkg = self.by_error[anissue][anerror]
                for line in pkg:
                    _, _, report, _, entity = line
                    if report == R_CONFIRMATION:
                        key = "Other"
                        if entity.is_parent:
                            key = "parent"
                        if entity.is_polo:
                            key = "polo"
                        if entity.is_synchro:
                            key = "synchro"
                        if entity.is_swimmer:
                            key = "swimmer"
                        if key in matrix:
                            matrix[key].append(entity)
                        else:
                            matrix[key] = [entity]

        res = ""
        for msg in matrix:
            res += f"{msg}: \n"
            for entity in matrix[msg]:
                res += f"{entity.email}; "
            res += "\n\n"
        return res


def create_dict(xdict, key1, key2, val1, val2, rpt, rev, entity):
    """Create 2 dimensiaonal dictionary."""
    # pylint: disable=too-many-arguments
    if key1 not in xdict:
        xdict[key1] = {}
    if key2 not in xdict[key1]:
        xdict[key1][key2] = []
    xdict[key1][key2].append([val1, val2, rpt, rev, entity])


def print_dict(xdict, reports):
    """Print all issues."""
    # pylint: disable=too-many-nested-blocks
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-locals
    res = ""

    for key1 in sorted(xdict.keys()):
        match = False
        to_print = f"{key1}:\n"
        for key2 in sorted(xdict[key1]):
            to_print += f"    {key2}"

            debug(f"PRINT ISSUE: {key1} / {key2}", 6)
            inner_match = False
            first = True
            length = len(xdict[key1][key2])

            for xissue in sorted(
                xdict[key1][key2], key=lambda x: "None" if x[0] is None else x[0]
            ):
                val1, val2, rpt, rev, _ = xissue
                if (first and rev) or (first and (length > 1)):
                    to_print += "\n"
                    first = False
                if (reports and (rpt in reports)) or (reports is None):
                    match = True
                if val1:
                    inner_match = True
                    if rev:
                        spacer = "        "
                    else:
                        val1 = f" ({val1})"
                        if length > 1:
                            spacer = "        "
                        else:
                            spacer = ""
                    if val2:
                        val2 = f" ({val2})"
                    to_print += f"{spacer}{val1}{val2}\n"
            if inner_match is False:
                to_print += "\n"

        if match:
            res += to_print
            res += "\n"

    return res



"""Read and process CSV Files."""
import csv
import datetime
import ntpath
import re

from scm_helper.config import (
    A_ASA_CATEGORY,
    A_ASA_NUMBER,
    A_DOB,
    A_FIRSTNAME,
    A_KNOWNAS,
    A_LASTNAME,
    C_CHECK_SE_NUMBER,
    C_DOB_FORMAT,
    C_FILES,
    C_IGNORE_GROUP,
    C_MAPPING,
    SCM_DATE_FORMAT,
    get_config,
)
from scm_helper.files import Files
from scm_helper.notify import notify

CAT_RE = re.compile(r"\d")


class Csv(Files):
    """CSV file handling."""

    def __init__(self):
        """Initilaise CSV."""
        super().__init__()
        self._csv = []
        self.by_name = {}
        self.by_knownas = {}
        self.cfg_file = None

    def readfile(self, file, scm):
        """Read CSV file."""
        self._filename = file
        self._scm = scm

        self.cfg_file = ntpath.basename(file)

        notify(f"Reading {self.cfg_file}...\n")

        cfg_dob_format = get_config(
            scm, C_FILES, self.cfg_file, C_MAPPING, C_DOB_FORMAT
        )
        if cfg_dob_format is None:
            cfg_dob_format = SCM_DATE_FORMAT

        cfg_dob = get_config(scm, C_FILES, self.cfg_file, C_MAPPING, A_DOB)
        if cfg_dob is None:
            cfg_dob = A_DOB

        cfg_cat = get_config(scm, C_FILES, self.cfg_file, C_MAPPING, A_ASA_CATEGORY)
        if cfg_cat is None:
            cfg_cat = A_ASA_CATEGORY

        try:
            count = 0
            with open(file, newline="", encoding="utf-8-sig") as csvfile:
                csv_reader = csv.DictReader(csvfile)

                for row in csv_reader:
                    count += 1
                    if count == 0:
                        continue
                    self._csv.append(row)

                    # Fix DOB
                    if cfg_dob in row:
                        try:
                            row[cfg_dob] = datetime.datetime.strptime(
                                row[cfg_dob], cfg_dob_format
                            )
                        except ValueError as error:
                            notify(f"Date format error in CSV:\n{error}\n")
                            return False
                    if cfg_cat in row:
                        cat = CAT_RE.search(row[cfg_cat])
                        if cat:
                            row[cfg_cat] = cat.group(0)

            notify(f"Read {file}...\n")
            return True

        except EnvironmentError:
            notify(f"Cannot open csv file: {file}\n")
            return False

        except csv.Error as error:
            notify(f"Error in csv file: {file}\n{error}\n")
            return False

        notify("Unknown error\n")
        return False

    def analyse(self, scm):
        """Analyse the CSV."""
        # Yes, its complicated...
        # pylint: disable=too-many-locals
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-statements

        cfg = get_config(scm, C_FILES, self.cfg_file)
        if cfg is None:
            # ODD structure to keep line length in black limit
            prefix = "No configuration for "
            postfix = ", name matching only & default headers\n"
            notify(f"{prefix}{self._filename}{postfix}")

            cfg_first = A_FIRSTNAME
            cfg_last = A_LASTNAME
            cfg_knownas = A_KNOWNAS
            cfg_asa = A_ASA_NUMBER
            cfg_cat = A_ASA_CATEGORY
            cfg_dob = A_DOB
        else:
            cfg_first = get_config(scm, C_FILES, self.cfg_file, C_MAPPING, A_FIRSTNAME)
            cfg_last = get_config(scm, C_FILES, self.cfg_file, C_MAPPING, A_LASTNAME)
            cfg_knownas = get_config(scm, C_FILES, self.cfg_file, C_MAPPING, A_KNOWNAS)
            cfg_asa = get_config(scm, C_FILES, self.cfg_file, C_MAPPING, A_ASA_NUMBER)
            cfg_cat = get_config(scm, C_FILES, self.cfg_file, C_MAPPING, A_ASA_CATEGORY)
            cfg_dob = get_config(scm, C_FILES, self.cfg_file, C_MAPPING, A_DOB)

        cfg_check_se = get_config(scm, C_FILES, self.cfg_file, C_CHECK_SE_NUMBER)

        members = self._scm.members

        for row in self._csv:
            name = f"{row[cfg_first]} {row[cfg_last]}"
            self.by_name[name] = row
            if cfg_knownas:
                knownas = f"{row[cfg_knownas]} {row[cfg_last]}"
                self.by_knownas[knownas] = row
            else:
                knownas = "XYZXYZXYZ"  # will not match

            member = members.find_member(name)
            if member is None:
                member = members.find_member(knownas)
                if member is None:
                    self.file_error(name, "In file, not in SCM")
                    continue
            else:
                if member.is_active is False:
                    self.file_error(name, "In file & SCM but inactive")
                    continue

        for member in members.entities:
            if member.is_active is False:
                continue

            if not (member.is_swimmer or member.is_polo or member.is_synchro):
                continue

            if self.check_ignore(member):
                continue

            row = None
            good_name_match = False
            if member.name in self.by_name:
                row = self.by_name[member.name]
                good_name_match = True
            elif member.name in self.by_knownas:
                row = self.by_knownas[member.name]
                good_name_match = True
            elif member.knownas in self.by_name:
                row = self.by_name[member.knownas]
                good_name_match = True
            elif member.knownas in self.by_knownas:
                row = self.by_knownas[member.knownas]

            if row is None:
                msg = "In SCM but not in file"
                extra = f"(Group: {member.first_group})"
                self.file_error(member.name, msg, extra)
                continue

            if good_name_match is None:
                self.file_error(member.name, "Partial name match using knownas")

            errmsg = ""
            if cfg_check_se:
                if member.asa_number:
                    if int(member.asa_number) != int(row[cfg_asa]):
                        errmsg += "Different SE Number"
                else:
                    errmsg += "SE Missing in SCM"

                if member.asa_category:
                    if member.asa_category != row[cfg_cat]:
                        if errmsg:
                            errmsg += ", "
                        errmsg += "Different SE Category"
                else:
                    if errmsg:
                        errmsg += ", "
                    errmsg += "SE Category Missing in SCM"

                if member.dob:
                    if member.dob != row[cfg_dob]:
                        if errmsg:
                            errmsg += ", "
                        errmsg += "Different Date of Birth"
                else:
                    if errmsg:
                        errmsg += ", "
                    errmsg += "DOB Missing in SCM"

            if errmsg:
                self.file_error(member.name, errmsg)

    def check_ignore(self, member):
        """Return true if swimmer in ignore list."""
        cfg_ignore = get_config(self._scm, C_FILES, self.cfg_file, C_IGNORE_GROUP)
        if cfg_ignore:
            for ignore in cfg_ignore:
                if member.find_group(ignore):
                    return True
        return False



"""Read and process CSV Files."""
import ntpath


class Files:
    """General file handling."""

    def __init__(self):
        """Initiase."""
        self._filename = None
        self._scm = []
        self._error = {}

    def file_error(self, name, error, extra=None):
        """Log an error in the file."""
        if error in self._error:
            self._error[error].append([name, extra])
        else:
            self._error[error] = []
            self._error[error].append([name, extra])

    def print_errors(self):
        """Print errors."""
        fname = ntpath.basename(self._filename)
        output = ""

        for error in self._error:
            output += f"{error}:\n"
            for name, extra in sorted(self._error[error]):
                output += f"   {name}"
                if extra:
                    output += f" {extra}"
                output += "\n"

        if output == "":
            output = f"\nNo issues in {fname}\n"
        else:
            output = f"\nIssues in {fname}\n" + output

        return output



"""Null Encryption stuff for ipad."""
import os.path
from pathlib import Path

from scm_helper.config import CONFIG_DIR
from scm_helper.notify import interact, notify

WRITE_BINARY = "w"
READ_BINARY = "r"

# pylint: disable=unused-argument
# pylint: disable=no-self-use


class Crypto:
    """Encryption class copied from crypto - but does nothing."""

    def __init__(self, salt, password):
        """Initialise."""
        self.salt = salt

    def encrypt_file(self, name, data):
        """Encrypt file."""
        notify("Not implemented on iPad\n")

    def decrypt_file(self, name, xdate):
        """Decrypt file."""
        notify("Not implemented on iPad\n")

    def encrypt_backup(self, name, data):
        """Encrypt file."""
        notify("Not implemented on iPad\n")

    def decrypt_backup(self, name, xdate):
        """Decrypt file."""
        notify("Not implemented on iPad\n")
        return False

    def read_key(self, filename):
        """Read API key."""
        home = str(Path.home())
        filename = os.path.join(home, CONFIG_DIR, filename)

        if not os.path.exists(filename):
            return self.get_key(filename)
        try:
            with open(filename, READ_BINARY) as file:
                data = file.read()
            file.close()
            return data.strip()

        except OSError as error:
            notify(f"Cannot open key file: {error}\n")
            return None

    def get_key(self, filename):
        """Encrypt file."""
        apikey = interact("No SCM API keyfile, creating one.  Enter API key: ")

        try:
            with open(filename, WRITE_BINARY) as file:
                file.write(apikey)
            file.close()

            notify(f"Generate encrypted keyfille {filename}\n")
            return apikey

        except OSError as error:
            notify(f"Cannot open/write key file: {error}\n")
            return False

    def read_email_password(self, filename):
        """Read email password."""
        notify("Not implemented on iPad\n")
        return False

    def get_email_password(self, filename):
        """Encrypt Password file."""
        notify("Not implemented on iPad\n")
        return False



"""Default config file."""

import os
from pathlib import Path

from scm_helper.config import CONFIG_DIR, CONFIG_FILE, FILE_WRITE
from scm_helper.notify import interact, interact_yesno, notify


def create_default_config():
    """Create a default config file."""
    home = str(Path.home())

    cfg = os.path.join(home, CONFIG_DIR)
    cfg_file = os.path.join(home, CONFIG_DIR, CONFIG_FILE)

    notify("Welcome to SCM Helper.\nStarting configuration for first time use..\n")

    msg = "Please enter your Swimming Club Name: "
    clubname = interact(msg)

    msg = f"Create initial configuration file:\n   {cfg_file}\nfor '{clubname}'"
    if interact_yesno(msg) is False:
        return False

    content = DEFAULT_CONFIG.replace("###CLUB_NAME###", clubname)

    try:
        if os.path.exists(cfg) is False:
            os.mkdir(cfg)

        with open(cfg_file, FILE_WRITE) as file:
            file.write(content)

        notify("Done.\n")

        return True

    except OSError as error:
        notify(f"Cannot create file: {error}\n")
        return False


DEFAULT_CONFIG = """
################################################################
# Configuration of SCM-Helper
#
# https://github.com/ColinRobbins/scm-helper/
#
# You will need to change these parameters
# to reflect the situation at your club.
# See above link for documentation and a full example.
#
# Some lines are commented out with "# "
# To reactivate, remove the '#' and one ' '.
# I.e. YAML files are very sensitive to the exact spacing.
# Errors will be created if you get the spacing wrong.
# Use spaces not tabs.
################################################################

# Your swimming club name.
# If you change this, you will not be able to access
# backups taken before the change.
club: "###CLUB_NAME###"


allow_update:  False   # Set to true if this tool is allowed to edit SCM
                       # Set to false if you don't trust the software to be reliable.


################################################################
# Email config.
# Only needed if using --email from command line
################################################################

# email:
#  smtp_server: smtp.server.address
#    smtp_port: 25
#    username:
#   send_to:
#   tls: False
#   password: "email.enc"       # Created when needed.

################################################################
# File input, for correlation of Swim England file
#
# File must have a header row.
# The mapping maps this headers to the required fields.
################################################################
files:
    "swim_england.csv":
        check_se_number: true
        mapping:
          Firstname: 'First Name'
          Lastname: 'Surname'
          KnownAs: 'Known As'
          ASANumber: 'Member ID'
          ASACategory: 'Category'
          DOB: 'Date of Birth'
          dob_format: '%d/%m/%Y'
#         ignore_group:  # Don't report errors if swimmer is in these groups
#           - 'Resignations'


#############################################
# Swim England Correlations
#############################################
swim_england:
    base_url: "https://www.swimmingresults.org/"
    check_url: "https://www.swimmingresults.org/membershipcheck/member_details.php?myiref="
    test_id: 516115

################################################################
# Configure Swimmer Checks
################################################################
swimmers:
    username:
        min_age: 17     # Min age to have a user name
    parent:
        mandatory: True # Must have a parent
        max_age: 17     # Age a parent is mandatory until
    confirmation_difference:
        verify: False    # if there is > 3 month difference in confirm date
                        # between parent and child, report an issue.
    absence:
        time:  182      # Warn if not seen at swimming for this period of time

################################################################
# Configure Parent Checks
################################################################
parents:
    age:
        min_age: 17   # min age to be a parent
        child: 21     # Age at which child should not longer have a parent
    login:
        mandatory: True         # must have a login to SCM

################################################################
# Configure Member Checks
################################################################
members:
    confirmation:
        expiry: 365             # Warn if confirmation more than this many day old.
        align_quarter: True     # Align expiry to calendar quarter
                                # So reminders go in batches.
    dbs:
        expiry: 60      # Days warning prior to expiry.
    newstarter:
        grace:  90      # Don't give errors for 90 days,
                        # giving admin time to sort it all out
    inactive:
        time: 365       # Warn if member in inactive state for this many days

################################################################
# Configure Coaches
################################################################
coaches:
    role:
        mandatory: False # All coaches, must be in a coach role

################################################################
# Configure Roles
################################################################
roles:
    volunteer:
        mandatory: True       # if user in a role, the volunteer flag must be set
    login:
        unused: 180           # Error if not used in 180 days

# Examples...
#   role:
#     "Coaches":
#         check_permissions: False
#         is_coach: True                # Should have "Is A Coach" ticked.
#     "Register Taker":
#         check_restrictions: True        # Check they have session restrictions

jobtitle:
    ignore:
        - "Vice President"      # If VP, don't need to be a committee member.

################################################################
# Configure Groups
#
# Check this members of the group are in a relevant swimming session
################################################################
# groups:
#  priority:    # If a user is in multiple groups
                # these take priority (in this order) when printing
#    - 'Water Polo'
#    - 'Masters'

#   group:
#     'Membership Only':
#         no_sessions: true
#     'Senior Development':
#         session:
#           - Senior Development'
#         type: swimmer
#     'Masters':
#         sessions:
#           - 'Masters'
#         no_session_allowed:
#           - 'Life Members and Past Presidents'
#           - 'Membership Only'
#         unique: false
#         min_age: 17
#         type: swimmer
#     'Life Members and Past Presidents':
#         ignore_unknown: true    # Don't raise issue if not a swimmer, parent or coach.
#         unique: false
#     'Team Manager':
#         check_dbs: true
#         type: volunteer       # Must be of this type
#         unique: false
#     'Club Timekeeper':
#         unique: false
#         type: volunteer
#     'Resignations':
#         ignore_swimmer: true
#
################################################################
# Configure Sessions
#
# Check for attendance and correlation with group memberhip.
# Sessions not listed will not be checked for group correlation.
################################################################
sessions:
  absence:  120  # Number of days allowed to miss a session
  register: 60   # Alert if register not taken for this many days
#   session:
#     'Junior Squad':
#         groups:
#           - 'Junior Squad'
#     'Masters':
#         groups:
#           - 'Masters'
#     'Transition':
#         groups:
#           - 'Tadpoles'
#           - 'Development'
#           - 'Transition'
#         ignore_attendance: True
#     'Starts and Turns':
#         ignore_attendance: True

#############################################
# Codes of Conduct
#############################################

# conduct:
#    "Code of Conduct for Coaches":
#       types:
#         - "coach"
#     "Code of Conduct for Swimmers and Water Polo Players":
#       types:
#         - "swimmer"
#       ignore_group:
#         - "Tadpoles"

################################################################
# Facebook Correlations
################################################################

# facebook:
#   files:
#     - "Nottingham Leander Swimmers.html"
#   groups:
#     - https://www.facebook.com/groups/NottinghamLeander/
#     - https://www.facebook.com/groups/LeanderMasters/

################################################################
# Lists
#
# Configure the criteria for auto-generated lists
#
################################################################
lists:
    suffix: " (Generated)"  # Use to identity generated lists
    edit: False             # Allow script to modify generated lists
    confirmation: False     # Generate lists of non-confirmed and
                            # expired confirmation members
#    conduct:               # Create a list of members who have not signed the code of conduct
#      - "Code of Conduct Name"

#     list:
#        "Swimmer: Development Only":
#             group: "Development"
#             unique: True                # ONLY the group, no others
#             allow_group: "Water Polo"   # If unique is set, allow this too
#        "Swimmers: 17 and under on Dec 31":
#            type: "swimmer"
#            max_age_eoy: 17
#        "Water Polo: Men: 16 and over":
#            group: "Water Polo"
#            min_age: 16
#            gender: "male"

################################################################
# Entity Types
#
# These are reflected by the "Is a Coach"  etc
# Boxes in a members details page
################################################################

types:
    synchro:
        name: "Synchro"
        check_se_number: False
        parents: False              # Don't check parents, overrides Mandatory above.
    waterpolo:
        name: "Water Polo"
#         groups:                   # Specify if they must be members
#           - "Water Polo"          # of a specific group
    volunteer:
        ignore_coach: True          # If coach, do not need to be a volunteer
        ignore_committee: True      # If committee, do not need to be a volunteer
#         groups:
#           - "Team Manager"
#           - "Club Timekeeper"
#           - "Licensed Officials"
    committee:
        jobtitle: True              # must have a job title

################################################################
# Issue handling
#
# Override the default message for an issue.
# Set ignore: true to ignore the message.
#
# For a list of issue codes see:
# https://github.com/ColinRobbins/scm-helper/wiki/Configuration-Issues
#
################################################################

issues:
    E_CONFIRMATION_EXPIRED:
        ignore_error: False
        message: "Confirmation expired"

##################################################
# Selenium
# Used for Swim England and Facebook checking
##################################################

#selenium:
#    browser: "Chrome"
#    web_driver: "C:\\Program Files (x86)\\Python37-32\\Scripts\\chromedriver.exe"


##################################################
# Records
##################################################

#records:
#    relay: false
#    age_eoy: true   # Calculate age at EOY, not age from SCM export file (where possible)
#    verify: true  # Check a club member at time of swim (where possible)
#    se_only: false



# Debug level, set to 0 for no debug info
debug_level: 0


"""



"""Send email."""
import smtplib

from scm_helper.config import (
    C_EMAIL,
    C_PASSWORD,
    C_SEND_TO,
    C_SMTP_PORT,
    C_SMTP_SERVER,
    C_TLS,
    C_USERNAME,
    O_TO,
    get_config,
)
from scm_helper.notify import notify


def send_email(scm, text, subject):
    """Send an email."""
    smtp_server = get_config(scm, C_EMAIL, C_SMTP_SERVER)
    smtp_port = get_config(scm, C_EMAIL, C_SMTP_PORT)
    username = get_config(scm, C_EMAIL, C_USERNAME)
    tls = get_config(scm, C_EMAIL, C_TLS)
    password_file = get_config(scm, C_EMAIL, C_PASSWORD)

    password = scm.crypto.read_key(password_file)
    if password is None:
        return False

    if scm.option(O_TO):
        send_to = scm.option(O_TO)
    else:
        send_to = get_config(scm, C_EMAIL, C_SEND_TO)

    recips = send_to.split(";")

    if subject is None:
        subject = "Message from Swim Club Manager"

    message = f"From: Swim Club Manager Helper <{username}>\n"
    message += f"To: <{send_to}>\n"
    message += f"Subject: {subject}\n\n"
    message += text

    try:
        email = smtplib.SMTP(smtp_server, smtp_port)
        if tls:
            email.starttls()
        email.login(username, password)
        email.sendmail(username, recips, message)
        notify("email sent.")

    except (smtplib.SMTPException, OSError) as error:
        notify(f"Error sending email\n{error}\n")
        return False

    return True
